******************************************************************************;
*  Copyright (c) 2005-2012 by SAS Institute, Inc., Cary, NC, USA.            *;
*  All rights reserved.                                                      *;
******************************************************************************;
*  NAME:        brm_export_ruleset.sas                                       *;
*  DESCRIPTION: This code is used to export ruleset to csv file.             *;
*                                                                            *;               
*  PARAMETERS:                                                               *;
*   rulesets=    The rulesets to exported.  This is a comma-separated list   *;
*                of ruleset_sk svalues, e.g. %str(10000, 10001).  If ALL is  *;
*                specified, then all rulsets will be exported.               *;
*                                                                            *;
*   folder_path= The folder of the rulesets to be exported.  All rulesets in *;
*                and under the folder path will be exported,                 *;
*                                                                            *;
*   csv=         The exported csv file, e.g. %str(c:\temp\ruleset.csv)       *;
*                                                                            *;
*   dbms=        dbms engine (default is ORACLE)                             *;
*                                                                            *;
*   dbms_path=   libname option (defaults to tokenized install value)        *;
*                                                                            *;
*   authdomain=  authentication domain (defaults to tokenized install value) *;
*                                                                            *;
*  PRODUCT:     Business Rule Management                                     *;
*  OWNER:       Jane Xue                                                     *;
******************************************************************************;
%macro brm_export_ruleset(rulesets=,csv=,folder_path=,
                          dbms=ORACLE,dbms_path=,authdomain=brm_db_auth);

    * Include some other useful macros we neeed ;
    %include "@sasmacro.dir@brm_obscount.sas";
    %include "@sasmacro.dir@brm_alloc.sas";
    %include "@sasmacro.dir@brm_release.sas";
    %include "@sasmacro.dir@brm_create_folder_path_dataset.sas";

	*Connect to DBMS use libname*;
    %brm_alloc(dbms=&dbms,dbms_path=&dbms_path,authdomain=&authdomain,access=readonly); 

	*Judge passed csv file*;
	%if ("&csv" eq "") %then %do;
		%put Please specify csv qualified csv file;
		%goto exit;
	%end;
    
	*Judge passwed rulesets*;
	%if ("&rulesets" eq "" and "&folder_path" eq "") %then %do;
		%put No rulesets exported!;
        %goto exit;
	%end;
	%else %if ((%upcase("&rulesets") eq "ALL") or 
               ("&folder_path" ne "")) %then %do; 
        %let rulesets=1;
	%end;
	%else %do;
		%let rulesets=%str(a.RULE_SET_SK in (&rulesets));
	%end;


    * create sub-components of export dataset ;
   proc sql noprint;
      create table ruleset_condition as
        select distinct 
           a.RULE_SET_SK,
           a.RULE_SET_NM,
           a.RULE_SET_SHORT_DESC as RULE_SET_DESC label='RULE_SET_DESC',
           b.VOCAB_NM,
           c.FOLDER_OBJECT_PATH as RULE_SET_PATH label='RULE_SET_PATH',
           d.RULE_NM,
           d.RULE_SHORT_DESC as RULE_DESC label='RULE_DESC',
           d.RULE_DESIGN_SEQ_NO as RULE_SEQ_NO label='RULE_SEQ_NO',
           d.CONDITIONAL_NM,
           i.VOCAB_TERM_NM as LHS_TERM label='LHS_TERM',
           e.RULE_COND_EXPR_TXT as EXPRESSION label='EXPRESSION',
           e.RULE_COND_DISPLAY_SEQ_NO+1 as EXPRESSION_ORDER label='EXPRESSION_ORDER',
           'CONDITION' as EXPRESSION_TYPE label='EXPRESSION_TYPE'        

        from
           brm.DCM_RULE_SET                 as a ,
           brm.DCM_VOCABULARY               as b,
           brm.DCM_FOLDER_OBJECT_ASSOC      as c,
           brm.DCM_RULE                     as d,
           brm.DCM_RULE_CONDITION           as e,
           brm.DCM_RULE_COND_X_VOCAB_TERM   as f,
           brm.DCM_VOCABULARY_TERM          as i

        where  
           a.RULE_SET_SK=d.RULE_SET_SK=c.FOLDER_OBJECT_SK  and
           upcase(c.FOLDER_OBJECT_TYPE_CD)='RULESET'       and
           a.VOCAB_SK=b.VOCAB_SK                           and 
           d.RULE_SK=e.RULE_SK                             and
           e.RULE_COND_SK=f.RULE_COND_SK                   and                         
           f.VOCAB_TERM_SK=i.VOCAB_TERM_SK                 and
           upcase(f.LHS_OPERAND_FLG)='Y'                   and
           &rulesets                                       
        order by rule_set_nm, rule_set_path, rule_seq_no, expression_order    ;      

      create table ruleset_action as
        select distinct
          a.RULE_SET_SK,
          a.RULE_SET_NM,
          a.RULE_SET_SHORT_DESC as RULE_SET_DESC label='RULE_SET_DESC',
          b.VOCAB_NM,
          c.FOLDER_OBJECT_PATH as RULE_SET_PATH label='RULE_SET_PATH',
          d.RULE_NM,
          d.RULE_SHORT_DESC as RULE_DESC label='RULE_DESC',
          d.RULE_DESIGN_SEQ_NO as RULE_SEQ_NO label='RULE_SEQ_NO',
          d.CONDITIONAL_NM,
          i.VOCAB_TERM_NM as LHS_TERM label='LHS_TERM',
          g.RULE_ACTION_EXPR_TXT as EXPRESSION label='EXPRESSION',
          g.RULE_ACTION_DISPLAY_SEQ_NO+1 as EXPRESSION_ORDER label='EXPRESSION_ORDER',
          'ACTION' as EXPRESSION_TYPE label='EXPRESSION_TYPE'            

        from 
          brm.DCM_RULE_SET                 as a ,
          brm.DCM_VOCABULARY               as b,
          brm.DCM_FOLDER_OBJECT_ASSOC      as c,
          brm.DCM_RULE                     as d,
          brm.DCM_RULE_ACTION              as g,
          brm.DCM_RULE_ACTION_X_VOCAB_TERM as h,
          brm.DCM_VOCABULARY_TERM          as i

        where  
          a.RULE_SET_SK=d.RULE_SET_SK=c.FOLDER_OBJECT_SK  and
          upcase(c.FOLDER_OBJECT_TYPE_CD)='RULESET'       and
          a.VOCAB_SK=b.VOCAB_SK                           and 
          d.RULE_SK=g.RULE_SK                             and                   
          g.RULE_ACTION_SK=h.RULE_ACTION_SK               and 
          h.VOCAB_TERM_SK=i.VOCAB_TERM_SK                 and
          upcase(h.LHS_OPERAND_FLG)='Y'                    and 
          &rulesets                                      
       order by rule_set_nm, rule_set_path, rule_seq_no, expression_order    ;      
    quit;


	* Update each  path to be name paths for rule_set_path and rule_flow_path ;
	* Build the path list dataset which we can join to ;
    %brm_create_folder_path_dataset;

	*Create dataset for export*;
	data rulesets ;
		set ruleset_condition ruleset_action;
		by rule_set_nm rule_set_path rule_seq_no expression_order EXPRESSION_TYPE  ;
	run;

	proc sort data=rulesets out=rulesets2; by rule_set_nm rule_set_path rule_seq_no descending EXPRESSION_TYPE expression_order  ;
	run;

	* Strip off the trailing object id for the rule set respectively for the current id paths. ;
	%brm_strip_object_path(DATASET=rulesets2,ORIGINAL_COLUMN=RULE_SET_PATH,NEW_COLUMN_NAME=RULE_SET_ID_FOLDER_PATH);
	* Join in the new paths to the dataset which are name paths ;
	%brm_add_folder_path(DATASET=rulesets2,COLUMN_NAME=RULE_SET_ID_FOLDER_PATH,PATH_COLUMN_NAME=RULE_SET_PATH_NAME);


	*Filter the exported vocabularies according to &folder_path *;
	%if ("&folder_path" eq "") %then %do;
		%let folder=1;
	%end;
	%else %do;
		%let likeExpression="&folder_path%";
		%let folder=%str(RULE_SET_PATH_NAME like &likeExpression);
	%end;

	* Replace the original columns with the values we want for path and drop the columns generated;
	data rulesets2 (drop=RULE_SET_PATH_NAME RULE_SET_ID_FOLDER_PATH count);
		set rulesets2;
		where &folder;
		by rule_set_nm rule_set_path rule_seq_no descending EXPRESSION_TYPE;
		length count 8;
		retain count 0;	
		RULE_SET_PATH=RULE_SET_PATH_NAME;
		if first.EXPRESSION_TYPE then do;
			count = 0;
		end;
		count = count + 1;
		EXPRESSION_ORDER = count;
    run;

	*Export dataset to csv file*;
	proc export data=work.rulesets2
      outfile="&csv"
      dbms=csv 
      replace;
	run;

%exit:
	*Delete unused temp tables*;
	 proc datasets lib=work memtype=data noprint;
       %if (%sysfunc(exist(work.ruleset_action)) eq 1) %then %do;
          DELETE ruleset_action;
       %end;
       %if (%sysfunc(exist(work.ruleset_condition)) eq 1) %then %do;
          DELETE ruleset_condition;
       %end;
	   %if (%sysfunc(exist(work.rulesets)) eq 1) %then %do;
          DELETE rulesets;
       %end;
	   %if (%sysfunc(exist(work.folderdata)) eq 1) %then %do;
          DELETE folderdata;
       %end;
       %if (%sysfunc(exist(work.fullpathfolderdata)) eq 1) %then %do;
          DELETE fullpathfolderdata;
       %end;
	   %if (%sysfunc(exist(work.rulesets2)) eq 1) %then %do;
          DELETE rulesets2;
       %end;
     run; 
     quit;  
          
	*deallocate BRM libname*;
    %brm_release(); 
 
%mend brm_export_ruleset;
