******************************************************************************;
*  Copyright (c) 2005-2012 by SAS Institute, Inc., Cary, NC, USA.            *;
*  All rights reserved.                                                      *;
******************************************************************************;
*  NAME:        brm_import_rule_flow.sas                                     *;
*  DESCRIPTION: This code is used to import ruleset from csv file to DBMS    *;
*  PARAMETER:   csv, the csv file, e.g. csv=%str(c:\temp\ruleset.csv)        *;
*               brm_ser, brm log on user                                     *;
*               reject, the csv file contains all reject rule flows          *;
*                       e.g., reject=%str(c:\temp\reject.csv)                *;
*  PRODUCT:     Business Rule Management                                     *;
*  NOTEs:       1. Please backup the database before run this code,          *;
*                  run this code may cause data lost                         *;
*               2. Please do not change RULE_FLOW_SK in csv if want to keep  *;
*                  the RULE_FLOW_SK, however you can delete RULE_FLOW_SK     *;
*               3. For FOLDER_OBJECT_PATH, if RULE_FLOW_SK have value,       *;
*                  then it is RULE_FlOW_PATH, if RULE_FLOW_SK not have value *;
*                  then it is FOLDER_PATH.                                   *;
*               4. For missng vlaue in CSV, for numeric field,               *;
*                  e.g., RULE_FLOW_SK, Please user period(.)                 *;
*                  For character missing value, please use blank             *;
*               5. If csv does not have RULE_FLOW_SK, then this code will    *;
*                  only add one ruleset node                                 *;
*  OWNER:       Jane Xue                                                     *;
******************************************************************************;
%macro brm_import_rule_flow(csv=,
         brm_user=&sysuserid,
         bypassLock=N,
         dbms=ORACLE,
         dbms_path=,
         authdomain=brm_db_auth,reject=);
	
    * Include some other useful macros we neeed ;
	%include "@sasmacro.dir@brm_alloc.sas";
	%include "@sasmacro.dir@brm_release.sas";
	%include "@sasmacro.dir@brm_obscount.sas";
	%include "@sasmacro.dir@brm_create_folder_path_dataset.sas";

	*Connect to DBMS use libname*;
    %brm_alloc(dbms=&dbms,dbms_path=&dbms_path,authdomain=&authdomain,access=write); 

    * check for import rule flow lock ;
	%if (%sysfunc(exist(brm.lock_import_rule_flow)) eq 1) %then %do;
       %if (%upcase("&bypassLock") eq "Y") %then %do;
         %put Deleting old lock table by force due to byPassLock= setting.;
	  	 proc sql noprint;
			drop table brm.lock_import_rule_flow;
		 quit;
	   %end;
       %else %do;
          proc sql noprint;
            select importUser, put(lockAcquiredDT,nldatm.) 
            into :importUser, :importWhen 
            from brm.lock_import_rule_flow;
          quit;
          %put Import aborted - %trim(&importUser) started a rule flow import at %trim(&importWhen) that has not completed; 
          %goto lockexit;
       %end;
    %end;

    * create the lock file for this import;
    proc sql noprint;
	   create table brm.lock_import_rule_flow as
          select "&brm_user" as importUser,
                 datetime() as lockAcquiredDT 
          from brm.dual;
    quit;
    run;


	*Verify the csv file exist*;
	%if ("&csv" eq "" ) %then %do;
	    %put ERROR: csv= parameter not specified;
	    %goto exit;
	%end;
	
	*Import csv file to dataset*;
	data WORK.CSV    ;
		%let _EFIERR_ = 0; /* set the ERROR detection macro variable */
		infile "&csv" delimiter = ',' MISSOVER DSD lrecl=32767 firstobs=2 ;
		informat RULE_FLOW_SK best12. ;
		informat RULE_FLOW_NM $100. ;
		informat RULE_FLOW_SHORT_DESC $256. ;
		informat RULE_SET_SECTION_CODE $20. ;
		informat RULE_FLOW_PATH $4000. ;
		informat RULE_SET_NM $100. ;
		informat RULE_SET_PATH $4000. ;
		informat VOCAB_NM $100. ;
		informat BY_TERM $32. ;		
		informat order best12. ;
		format RULE_FLOW_SK best12. ;
		format RULE_FLOW_NM $100. ;
		format RULE_FLOW_SHORT_DESC $256. ;
		format RULE_SET_SECTION_CODE $20. ;
		format RULE_FLOW_PATH $4000. ;
		format RULE_SET_NM $100. ;
		format RULE_SET_PATH $4000. ;
		format VOCAB_NM $100. ;
		format BY_TERM $1000. ;		
		format order best12. ;
		input
			RULE_FLOW_SK
			RULE_FLOW_NM $
			RULE_FLOW_SHORT_DESC $
			RULE_SET_SECTION_CODE $
			RULE_FLOW_PATH $
			RULE_SET_NM $
			RULE_SET_PATH $
			VOCAB_NM $
			BY_TERM $			
			order
		;
		if _ERROR_ then call symputx('_EFIERR_',1);  /* set ERROR detection macro variable */
	run;

	/*If there are no sk's then the column type gets defaulted to text which is wrong.
	  reset it to a number.*/
	%let populatedsk = -1;
	proc sql noprint;
		select count(*) into :populatedsk
		from csv where not missing(rule_flow_sk);
	quit;
	
	%if &populatedsk=0 %then %do;
		data csv (drop=RULE_FLOW_SK);
			set csv;
		run;
		data csv;
			length RULE_FLOW_SK 8;
			set csv;
			RULE_FLOW_SK=.;
		run;
	%end;

	*Set VOCAB_NM length to 100*;
	data csv;
		length VOCAB_NM $100;
		set csv;
	run;

	%if %brm_obscount(csv)=0 %then %do;
   	  %put ERROR: The csv does not have data, import will abort!;
	  %goto exit;
    %end;

	/* initialize the folder resolving datasets. */
	%brm_create_folder_path_dataset;

	/* Validate that the rule sets exist within their folder location stated. Also
	   resolve the rule set sk for everything found.  */
	proc sql noprint;
		create table ruleflow_2 as
		select rf.*, brs.rule_set_sk as rule_set_sk 
		from fullpathfolderdata fp,
		     csv rf,
			 brm.DCM_RULE_SET brs,
			 brm.DCM_FOLDER_OBJECT_ASSOC foa
		where
			fp.fsk = foa.folder_sk and
			rf.rule_set_path = fp.foldernamepath and
			brs.rule_set_nm = rf.rule_set_nm and
			foa.folder_object_type_cd = 'RULESET' and
			foa.folder_object_sk = brs.rule_set_sk
			;

	quit;

	/* create a table of affected rule flows. */
	proc sql noprint;
		create table missingrulesets as 
		select rule_flow_path, rule_flow_nm from
			csv c where c.rule_flow_nm not in (
				select rf.rule_flow_nm from ruleflow_2 rf
					where rf.rule_set_nm = c.rule_set_nm 
						and rf.rule_set_path = c.rule_set_path
						and rf.rule_flow_path = c.rule_flow_path);
	quit;

	/*Delete those rule flows from the ruleflow_2 data set*/
	proc sql noprint;
		delete from ruleflow_2 rf
					where rf.rule_flow_nm in (select c.rule_flow_nm from missingrulesets c where 
						rf.rule_flow_nm = c.rule_flow_nm  and rf.rule_flow_path = c.rule_flow_path);
	quit;

	%let ruleset=.;
	proc sql noprint;		
        select count(*) into :ruleset
		from   csv;
	quit;	

	%let ruleset_validated=.;
	proc sql noprint;		
        select count(*) into :ruleset_validated
		from   ruleflow_2;
	quit;	
	

	%if &ruleset > &ruleset_validated  %then %do;
		%put WARNING: The csv has rule flows that have invalid rule sets, these rule flows will not be imported.;
	%end;
	

    %if %brm_obscount(ruleflow_2)=0 %then %do;
   	  %put ERROR: The csv does not have valid data, import will abort!;
	  %goto exit;
    %end;

	/* Check that all the path for the rule flows exists. Anyone that doesn't throw it out.*/
	
	proc sql noprint;
	
    create table ruleflow_3 as
	select rf.*, fsk as parent_folder_sk
	from fullpathfolderdata fp,
	     ruleflow_2 rf
	where
		rf.rule_flow_path = fp.foldernamepath
		;

	update ruleflow_3 rf set rule_flow_sk = (
		select brf.rule_flow_sk
		from fullpathfolderdata fp,
	     brm.DCM_RULE_FLOW brf,
		 brm.DCM_FOLDER_OBJECT_ASSOC foa
	where
		fp.fsk = foa.folder_sk
		and brf.rule_flow_nm = rf.rule_flow_nm
		and rf.rule_flow_path = fp.foldernamepath
		and foa.folder_object_type_cd = 'RULEFLOW'
		and brf.rule_flow_sk = foa.folder_object_sk		
	);
	quit;

	%let folder=.;
    proc sql noprint;
   	   select count(*) into :folder
	   from ruleflow_2;
    quit;

	%let folder_validated=.;
    proc sql noprint;
   	   select count(*) into :folder_validated
	   from ruleflow_3;
    quit;
	
    %if &folder>&folder_validated %then %do;
   	  %put WARNING: The csv has rule flows with invalid RULE_FLOW_PATH, the rule flows with these paths will not be imported.;
    %end;

    
    %if %brm_obscount(ruleflow_3)=0 %then %do;
   	  %put ERROR: The csv does not have valid data, import will abort!;
	  %goto exit;
    %end;


    *Verify VOCAB_NM exit in DCM_VOCABULARY*;
	%let vocab=.;	
	proc sql noprint;
		select count(*) into :vocab 
		from ruleflow_3
		where VOCAB_NM not in (select VOCAB_NM from brm.DCM_VOCABULARY) and 
              not missing(BY_TERM);
	quit;

	%if &vocab>0 %then %do;
		%put WARNING: The csv contains VOCAB_NM that are not in DCM_VOCABULARY table, the related rule flows will not be imported.;
	%end;

	proc sql noprint;
		create table ruleflow_4 as
		select *
		from ruleflow_3
		where VOCAB_NM in (select VOCAB_NM from brm.DCM_VOCABULARY) or 
              missing(BY_TERM);
	quit;

	*Verify rule_flow_nm is valid SAS Name*;
	%let nos=%brm_obscount(ruleflow_4);
	proc sql noprint;
		select trim(RULE_FLOW_NM) into :ruleflow1 - :ruleflow&nos
 		from ruleflow_4;
	quit;

	%local valid_name; 
    %do i=1 %to &nos; 
		%let valid_name =%sysfunc(NVALID(%nrbquote(&&ruleflow&i),V7));
		%if &valid_name=0 %then %do;
			proc sql noprint;
				delete * from ruleflow_4 
				where %NRBQUOTE(trim(RULE_FLOW_NM))="&&ruleflow&i"; 
			quit;
			%put WARNING: The csv contains invalid rule flow names, these ruleflows will not be imported.;

		%end;
     %end;
	
	 *Verify RULE_SET_SECTION_CODE is valid section*;
	 %let section=.;
	 proc sql noprint;
	 	select count (*) into :section
		from ruleflow_4
		where upcase(RULE_SET_SECTION_CODE) not in ('INIT','GROUPSTART','MAIN','GROUPEND','FINAL');
	 quit;

	%if &section>0 %then %do;
		proc sql noprint;
			delete * 
			from ruleflow_4
			where RULE_FLOW_NM in (select RULE_FLOW_NM from ruleflow_4 where upcase(RULE_SET_SECTION_CODE) not in ('INIT','GROUPSTART','MAIN','GROUPEND','FINAL'));
		quit;
		
		%put WARNING: The csv contains rule flows that have invalid rule set section codes, these rule flows will not be imported.;

	%end;


	%if %brm_obscount(ruleflow_4)=0 %then %do;
   	  %put ERROR: The csv does not have valid data, import will abort!;
	  %goto exit;
    %end;
          /*
			validate that all rule sets exists within the same vocabulary for each rule flow.
			validate that the vocabulary matches the rule sets as well.
			*/
	proc sql noprint;
		create table ruleflow_rsv as
		select *
		from ruleflow_4 rf,
			brm.DCM_RULE_SET rs,
			 brm.DCM_VOCABULARY v
		where v.VOCAB_NM = rf.VOCAB_NM
              and rs.VOCAB_SK = v.VOCAB_SK
			  and rf.RULE_SET_SK = rs.RULE_SET_SK;
	quit;

	/* remove the entire rule flow for any rule set mismatch.*/
	proc sql noprint;
		create table ruleflow_deletes as select rf4.rule_flow_nm, rf4.rule_flow_path
		from 
			ruleflow_4 rf4
		where rf4.rule_flow_nm not in (select rfv.rule_flow_nm 
											from ruleflow_rsv rfv where 
											rfv.rule_flow_path = rf4.rule_flow_path and
											rfv.rule_flow_nm = rf4.rule_flow_nm and
											rfv.RULE_SET_SK = rf4.RULE_SET_SK);
	quit;

	%let rulesetvocab=;	
	proc sql noprint;
		select count(*) into :rulesetvocab 
		from ruleflow_deletes;
	quit;

	%if &rulesetvocab>0 %then %do;
		%put WARNING: The csv contains rule sets that are not using the referenced vocabulary, the related rule flows will not be imported.;

	%end;
	
	proc sql noprint;
		delete from 
			ruleflow_rsv rsv
		where rsv.rule_flow_nm in ( select rfv.rule_flow_nm 
											from ruleflow_deletes rfv where 
											rfv.rule_flow_path = rsv.rule_flow_path and
											rfv.rule_flow_nm = rsv.rule_flow_nm);
	quit;

	%if %brm_obscount(ruleflow_rsv)=0 %then %do;
   	  %put ERROR: The csv does not have valid data, import will abort!;
	  %goto exit;
    %end;
	
	/*Validate that no rule set is used more that once within a rule flow.*/
	proc sql noprint;
		create table ruleflow_rsd as
		select RULE_FLOW_NM, RULE_FLOW_PATH, RULE_SET_SK, count(RULE_SET_SK) as rscount
		from ruleflow_rsv rsv
		group by RULE_FLOW_NM, RULE_FLOW_PATH, RULE_SET_SK;
	quit;
	
	%let rulesetdup=.;	
	proc sql noprint;
		select count(*) into :rulesetdup 
		from ruleflow_rsd where rscount > 1;
	quit;

	
	%if &rulesetdup>0 %then %do;
		%put WARNING: The csv contains rule sets that repeated multiple times within a rule flow, the related rule flows will not be imported.;

	%end;

	proc sql noprint;
		delete from 
			ruleflow_rsv rsv
		where rsv.rule_flow_nm in (select rfv.rule_flow_nm 
											from ruleflow_rsd rfv where 
											rfv.rule_flow_path = rsv.rule_flow_path and
											rfv.rule_flow_nm = rsv.rule_flow_nm and 
											rscount > 1);
	quit;

	%if %brm_obscount(ruleflow_rsv)=0 %then %do;
   	  %put ERROR: The csv does not have valid data, import will abort!;
	  %goto exit;
    %end;
	

	*Verify byterm exist in DCM_VOCABULARY_TERM table*;	
	data byterm(keep=RULE_FLOW_SK VOCAB_NM BY_TERM count);
		set ruleflow_rsv;
		where not missing(BY_TERM);
		count=count(BY_TERM,',');			
	run;

	data byterm1(drop=count BY_TERM);
		length term $32;
		set byterm;
		where count=0;		
		term=BY_TERM;
	run;

	data byterm2;
		set byterm;
		where count>0;
	run;

	%if %brm_obscount(byterm2)>0 %then %do;
		proc sql noprint;
			select count(*), max(count) into 
	               :count,
				   :max
			from byterm2;
		quit;
		%let count=&count;
		%let max=&max;

		proc sql noprint;
			select BY_TERM,count,RULE_FLOW_SK,VOCAB_NM into 
	               :term1  - :term&count,
				   :count1 - :count&count,
				   :sk1    - :sk&count,
				   :vocab1 - :vocab&count
			from byterm2;
		quit;	
		
		data byterm3;			
			length term term1 - term&max $32 VOCAB_NM $100;;
			%do i=1 %to &count;
				RULE_FLOW_SK=&&sk&i;
				VOCAB_NM="&&vocab&i";
				%do j=1 %to &&count&i;				    
					term&j=scan("&&term&i",&j,',');
					term=scan("&&term&i",-1,',');				
				%end;
				output;
			%end;			
		run;

		data byterm30(keep=RULE_FLOW_SK VOCAB_NM term);
			set byterm3;						
		run;
	    
		%do i=1 %to &max;
			data byterm3&i(keep=RULE_FLOW_SK VOCAB_NM term);
				set byterm3;
				rename term&i=term;
			run;			
	        
			data byterm30;
				set byterm30 byterm3&i;				
			run;

			%if (%sysfunc(exist(byterm3&i)) eq 1) %then %do;
				proc sql noprint;
					drop table byterm3&i;
				quit;
			%end;
		%end;

		data temp_byterm;
			set byterm1 byterm30;			
		run;
	%end;
	%else %if %brm_obscount(byterm2)=0 %then %do;
		data temp_byterm;
			set byterm1;			
		run;
	%end;
	
	proc sql noprint;
		create table invalid_term as
		select *
		from temp_byterm
		where term not in (
              select VOCAB_TERM_NM 
              from brm.DCM_VOCABULARY_TERM   as a,
                   brm.DCM_VOCABULARY_ENTITY as b,
				   brm.DCM_VOCABULARY        as c
              where a.VOCAB_TERM_NM=temp_byterm.term    and
			        a.VOCAB_ENTITY_SK=b.VOCAB_ENTITY_SK and 
					b.VOCAB_SK=c.VOCAB_SK               and 
					c.VOCAB_NM=temp_byterm.VOCAB_NM
              );
	quit;

	%if %brm_obscount(invalid_term)>0 %then %do;
		%put WARNING: The csv has by terms that do not exist in DCM_VOCABULARY_TERM, the related rule flows will not be imported.;

	%end;

	proc sql noprint;
		create table ruleflow_5 as 
		select *
		from ruleflow_rsv
		where RULE_FLOW_SK not in (select RULE_FLOW_SK from invalid_term);
	quit;

	%if %brm_obscount(ruleflow_5)=0 %then %do;
		%put ERROR: csv does not have valid data, import will abort!;
		%goto exit;
	%end;

	*Verify the order is continuous*;
	
	
   proc sort data=ruleflow_5;
   	  by RULE_FLOW_NM RULE_FLOW_PATH order;
   run;

   data order;
   	  set ruleflow_5;
	  by RULE_FLOW_NM RULE_FLOW_PATH;
	  retain count;
      if first.RULE_FLOW_PATH then do;
	  	count=0;
	  end;
	  count=count+1;
   run;

   data order;
   	  set order;
	  if order=count then do;
	  	 valid_flg=1;
	  end;
	  else do;
	  	valid_flg=0;
	  end;
   run;

   data invalid_order;
   	  set order;
	  where valid_flg=0; 
   run;
   
   %if %brm_obscount(invalid_order)>0 %then %do;
		%put WARNING: The csv has rule sets with invalid ordering numbers for the rule flow, the related rule flows will not be imported.;

	%end;

    proc sql noprint;
		create table Ruleflow_6 as
		select *
		from Ruleflow_5 rf
		where RULE_FLOW_NM not in (select io.RULE_FLOW_NM from invalid_order io 
									where io.RULE_FLOW_NM = rf.RULE_FLOW_NM and
										io.RULE_FLOW_PATH = rf.RULE_FLOW_PATH);
	quit;

	%if %brm_obscount(Ruleflow_6)=0 %then %do;
		%put ERROR: The csv does not have valid data, import will abort! ;
		%goto exit;
	%end;

	
	*RULE_SET_SECTION_CODE is properly used*;
	data section(keep=RULE_FLOW_NM RULE_FLOW_PATH order RULE_SET_SECTION_CODE) ;
		set Ruleflow_6;
	run;

	proc sort data=section;
		by RULE_FLOW_NM RULE_FLOW_PATH order;
	run;

	data init;
		set section;
		where upcase(RULE_SET_SECTION_CODE)='INIT';
	run;

	data main;
		set section;
		where upcase(RULE_SET_SECTION_CODE)='MAIN';
	run;

	data start;
		set section;
		where upcase(RULE_SET_SECTION_CODE)='GROUPSTART';
	run;

	data end;
		set section;
		where upcase(RULE_SET_SECTION_CODE)='GROUPEND';
	run;

	data final;		
		set section;
		where upcase(RULE_SET_SECTION_CODE)='FINAL';
	run;	
	
	proc sql noprint;
		create table invalid_init as
		select RULE_FLOW_NM, RULE_FLOW_PATH 
        from init
		where RULE_FLOW_NM in (select RULE_FLOW_NM from main  where RULE_FLOW_NM=init.RULE_FLOW_NM and RULE_FLOW_PATH=init.RULE_FLOW_PATH and order<init.order)  or
              RULE_FLOW_NM in (select RULE_FLOW_NM from start where RULE_FLOW_NM=init.RULE_FLOW_NM and RULE_FLOW_PATH=init.RULE_FLOW_PATH and order<init.order)  or
              RULE_FLOW_NM in (select RULE_FLOW_NM from end   where RULE_FLOW_NM=init.RULE_FLOW_NM and RULE_FLOW_PATH=init.RULE_FLOW_PATH and order<init.order)  or
			  RULE_FLOW_NM in (select RULE_FLOW_NM from final where RULE_FLOW_NM=init.RULE_FLOW_NM and RULE_FLOW_PATH=init.RULE_FLOW_PATH and order<init.order);
	quit;

	proc sql noprint;
		create table invalid_start as
		select RULE_FLOW_NM, RULE_FLOW_PATH  
        from start
		where RULE_FLOW_NM in (select RULE_FLOW_NM from main  where RULE_FLOW_NM=start.RULE_FLOW_NM and RULE_FLOW_PATH=start.RULE_FLOW_PATH and order<start.order)  or              
              RULE_FLOW_NM in (select RULE_FLOW_NM from end   where RULE_FLOW_NM=start.RULE_FLOW_NM and RULE_FLOW_PATH=start.RULE_FLOW_PATH and order<start.order)  or
			  RULE_FLOW_NM in (select RULE_FLOW_NM from final where RULE_FLOW_NM=start.RULE_FLOW_NM and RULE_FLOW_PATH=start.RULE_FLOW_PATH and order<start.order);
	quit;

	proc sql noprint;
		create table invalid_main as
		select RULE_FLOW_NM, RULE_FLOW_PATH  
        from main
		where RULE_FLOW_NM in (select RULE_FLOW_NM from end   where RULE_FLOW_NM=main.RULE_FLOW_NM and RULE_FLOW_PATH=main.RULE_FLOW_PATH and order<main.order)  or
			  RULE_FLOW_NM in (select RULE_FLOW_NM from final where RULE_FLOW_NM=main.RULE_FLOW_NM and RULE_FLOW_PATH=main.RULE_FLOW_PATH and order<main.order);
	quit;

	proc sql noprint;
		create table invalid_end as
		select RULE_FLOW_NM, RULE_FLOW_PATH  
        from end
		where order in (select order from final where RULE_FLOW_NM=end.RULE_FLOW_NM and RULE_FLOW_PATH=end.RULE_FLOW_PATH and order<end.order);
	quit;

	data invalid_section;
		set invalid_init invalid_start invalid_main invalid_end;
	run;

	%if %brm_obscount( invalid_section)>0 %then %do;
		%put WARNING: The csv has invalid RULE_SET_SECTION_CODE for some rule sets, the related rule flows will not be imported.;

	%end;

	proc sql noprint;
		create table Ruleflow_7 as
		select *
		from Ruleflow_6 rf
		where RULE_FLOW_NM not in (select io.RULE_FLOW_NM from invalid_section io 
									where io.RULE_FLOW_NM = rf.RULE_FLOW_NM and
										io.RULE_FLOW_PATH = rf.RULE_FLOW_PATH);
	quit;

	%if %brm_obscount( Ruleflow_7)=0 %then %do;
		%put ERROR: The csv does not have valid data, import will abort!;
		%goto exit;
	%end;

	*Add input and output nodes to csv file*;
	proc sql noprint;
		create table input as 
		select distinct 
			RULE_FLOW_SK,
			RULE_FLOW_NM,
			RULE_FLOW_PATH,
			RULE_FLOW_SHORT_DESC,
			parent_folder_sk
		from   ruleflow_7;
	quit;


	data ruleflow_7;
		length NODE_TYPE_CD $20;
		set ruleflow_7;
		NODE_TYPE_CD='RULESET';
	run;

	data input;
		length NODE_TYPE_CD $20;
		length ORDER 8;
		set input;
		NODE_TYPE_CD='INPUT';
		ORDER=.;
		output;
		NODE_TYPE_CD='OUTPUT';
		ORDER=.;
		output;
	run;

	/* Update the order number for each input and output node. */
	proc sql noprint;
		update input i set ORDER = (select min(order) - 1 from ruleflow_7 rf where 
										rf.rule_flow_nm = i.rule_flow_nm and 
										rf.rule_flow_path = i.rule_flow_path) where NODE_TYPE_CD = 'INPUT';
										
		update input i set ORDER = (select max(order) + 1 from ruleflow_7 rf where 
										rf.rule_flow_nm = i.rule_flow_nm and 
										rf.rule_flow_path = i.rule_flow_path) where NODE_TYPE_CD = 'OUTPUT';
	quit;

	proc sort data=ruleflow_7;
		by RULE_FLOW_SK
		   RULE_FLOW_NM
           RULE_FLOW_PATH
           ORDER;
	run;

	proc sort data=input;
		by RULE_FLOW_SK
		   RULE_FLOW_NM
           RULE_FLOW_PATH
           ORDER;
	run;
	
	data ruleflow;
		set ruleflow_7 input;
		by RULE_FLOW_SK RULE_FLOW_NM RULE_FLOW_PATH ORDER;
        NODE_OBJECT_SK=rule_set_sk; /* In our validation we looked up the ids and added them to the table already.*/
	run;	

	*update NODE_OBJECT_SK to RUEL_SET_SK for inputoutput nodes. This just ensures they are set to 0 *;
	proc sql noprint;
		update ruleflow 
		set NODE_OBJECT_SK=0
		where upcase(NODE_TYPE_CD)='INPUT' or upcase(NODE_TYPE_CD)='OUTPUT';
	quit;
	
	*Divide data into update and create*;
	data create_ruleflow;
		set ruleflow;
		where missing(RULE_FLOW_SK);
		sk=.;
		user="&brm_user";
	run;

	data update_ruleflow;
		set ruleflow;
		where not missing(RULE_FLOW_SK);
		sk=.;
		user="&brm_user";
	run;

	*First, reserve an history revision number by creating a record within the history table. 
	use the timestamp within the envers history table for all create/modify dates for the rule set data. *;
	proc sql noprint;
		 connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);;
		 select seqval into :version_no from connection to oracle (
		 select ENVERS_HST_SK_SEQ.nextval as seqval
		 from dual 
	 );
	quit;		 

	*Insert a record into DCM_ENVERS_INFO table*; 
	filename ver temp;
	data _null_;
		file ver;
		put 'proc sql noprint;';
		put '   connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);';
		put '   execute(Insert into DCM_ENVERS_INFO (ENVERS_HST_SK,ENVERS_HST_DTTM)';
		put "   values (&version_no.,sysdate)) by oracle;";
		put '   execute(commit) by oracle;';
		put 'quit;';
	run;
	%include ver;

	%if (%sysfunc(fileref(ver)) eq 0) %then %do;
		filename ver clear;
	%end;  

	*Update temp history table ENVERS_CHG_TYPE_CD from 0 to 2*;
	%macro brm_update_history_table(table=);
	filename update temp;
	data _null_;
		file update LRECL=32767;								
		put 'proc sql noprint;';
		put '	connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);';	
		put	"   execute (update &table set ENVERS_CHG_TYPE_CD=2) by oracle;";			   
		put '   execute(commit) by oracle;';				
		put '   disconnect from oracle;';	
		put 'quit;';
	run;

	%include update;

	%if (%sysfunc(fileref(update)) eq 0) %then %do;
	  filename update clear;
	%end;
	%mend brm_update_history_table;	

	*Create temp  table and import temp table records into DBMS*;
	*If RULE_SET_SK exist in csv file, then use update_ruleset table, if not exit, then user create_ruleset table*;	
	%macro brm_import_table(update=);
		 %if "&update"="Y" %then %do;
		  %let data=%str(work.update_ruleflow);
		  %let no=1;
		  %let create_dttm=%str(c.RULE_FLOW_CREATE_DTTM);
		  %let mod_dttm=%str(b.ENVERS_HST_DTTM   as RULE_FLOW_MOD_DTTM);
		  %let where =%str(a.RULE_FLOW_SK=c.RULE_FLOW_SK);	  
		  %let from=%str(,brm.DCM_RULE_FLOW as c);
		%end;
		%else %if "&update"="N" %then %do;
		  %let data=%str(work.create_ruleflow);
		  %let no=0;
		  %let create_dttm=%str(b.ENVERS_HST_DTTM   as RULE_FLOW_CREATE_DTTM);
		  %let mod_dttm=%str(b.ENVERS_HST_DTTM   as RULE_FLOW_MOD_DTTM);
		  %let where=1;
		  %let from=;	 
		%end;

		*Create temp ruleset and ruleset_hst tables*;
		%if (%sysfunc(exist(brm.temp_ruleflow)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_ruleflow;
			quit;
		%end;
		
		proc sql noprint;
			create table brm.temp_ruleflow as
			select distinct 
				   a.RULE_FLOW_SK,
				   a.RULE_FLOW_NM,
				   a.RULE_FLOW_SHORT_DESC,
				   a.RULE_FLOW_PATH,
				   a.user as RULE_FLOW_CREATE_USER_NM,
				   a.user as RULE_FLOW_MOD_USER_NM,
				   &create_dttm,
				   &mod_dttm

			from   &data as a,
				   brm.DCM_ENVERS_INFO as b
				   &from

			where  b.ENVERS_HST_SK=&version_no  and
			       &where;			   

		quit;		

		*If RULE_FLOW_SK exit in csv, then update DCM_RULE_FLOW table with temp_ruleflow*;
		%if "&update"="Y" %then %do;  
			proc sql noprint;
			   update brm.DCM_RULE_FLOW 
			   set RULE_FLOW_NM=(select RULE_FLOW_NM 
				                  from brm.temp_ruleflow
				                  where DCM_RULE_FLOW.RULE_FLOW_SK=temp_ruleflow.RULE_FLOW_SK),
				        
				    RULE_FLOW_SHORT_DESC=(select RULE_FLOW_SHORT_DESC 
				                          from brm.temp_ruleflow 
				                          where DCM_RULE_FLOW.RULE_FLOW_SK=temp_ruleflow.RULE_FLOW_SK),

				   
				     RULE_FLOW_MOD_USER_NM=(select RULE_FLOW_MOD_USER_NM 
				                            from brm.temp_ruleflow   
				                            where DCM_RULE_FLOW.RULE_FLOW_SK=temp_ruleflow.RULE_FLOW_SK),

				     RULE_FLOW_MOD_DTTM=(select RULE_FLOW_MOD_DTTM 
				                           from brm.temp_ruleflow 
				                           where DCM_RULE_FLOW.RULE_FLOW_SK=temp_ruleflow.RULE_FLOW_SK)

			     where DCM_RULE_FLOW.RULE_FLOW_SK = (select RULE_FLOW_SK from brm.temp_ruleflow  where DCM_RULE_FLOW.RULE_FLOW_SK=temp_ruleflow.RULE_FLOW_SK);
				                 
						  
			quit;
        %end;
		*If RULE_FLOW_SK does not exit in csv, then use RULE_FLOW_SK_SEQ.nextval as RULE_FLOW_SK,
		*and register the ruleflow in DCM_FOLDER_OBJECT_ASSOC*;
		%if "&update"="N" %then %do;

		  *Update RULE_FLOW_SK with RULE_FLOW_SK_SEQ.nextval *;
		  proc sql noprint;
		     connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);
		     execute (update temp_ruleflow set RULE_FLOW_SK=RULE_FLOW_SK_SEQ.nextval) by oracle; 
		     execute (commit) by oracle;
			 disconnect from oracle;
		  quit;

		  /* Update &data with the rule flow sk's generated for lower steps (specifically by terms)*/
			proc sql noprint;		   
				update &data a set RULE_FLOW_SK = (select b.RULE_FLOW_SK from brm.temp_ruleflow as b
												where strip(a.RULE_FLOW_PATH) = strip(b.RULE_FLOW_PATH) and
			                                                a.RULE_FLOW_NM = b.RULE_FLOW_NM ) 
											where missing(RULE_FLOW_SK) = 1;
		  	quit;

	      *Create temp table prepare for register rulesets to DCM_FOLDER_OBJECT_ASSOC*;
		proc sql noprint;		   
		  	create table temp_obj_asoc as 
			select distinct
				     a.RULE_FLOW_PATH,
				     b.RULE_FLOW_SK as FOLDER_OBJECT_SK,
				     a.parent_folder_sk as FOLDER_SK, 
					 (select c.path_sk from fullpathfolderdata c where strip(a.RULE_FLOW_PATH) = strip(c.foldernamepath)) as FOLDER_SK_PATH,
					 (select max(FOLDER_OBJECT_ORDER_NO) + 1 from brm.DCM_FOLDER_OBJECT_ASSOC where FOLDER_SK = a.parent_folder_sk) as FOLDER_OBJECT_ORDER_NO
					 
			from     &data                       as a,
					 brm.temp_ruleflow           as b
				     
	        where   strip(a.RULE_FLOW_PATH) = strip(b.RULE_FLOW_PATH) and
			        a.RULE_FLOW_NM = b.RULE_FLOW_NM and 
					missing(a.parent_folder_sk) = 0;

			update temp_obj_asoc set FOLDER_OBJECT_ORDER_NO = 0 where missing(FOLDER_OBJECT_ORDER_NO) = 1;
		  quit;

		  data temp_obj_asoc;
		    set temp_obj_asoc;
		  	FOLDER_OBJECT_PATH=strip(FOLDER_SK_PATH)||'/'||strip(FOLDER_OBJECT_SK);
		  run;

		  %if (%sysfunc(exist(brm.temp_obj_asoc)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_obj_asoc;
			quit;
		  %end; 	

		  data brm.temp_obj_asoc;
		  	set temp_obj_asoc;
			FOLDER_OBJECT_ASSOC_SK=.;			
			FOLDER_OBJECT_TYPE_CD='RULEFLOW';
		  run;
		 
		  *Update FOLDER_OBJECT_ASSOC_SK with FOLDER_OBJECT_ASSOC_SK_SEQ.nextval, and register rulesets records to DCM_FOLDER_OBJECT_ASSOC*;
		  proc sql noprint;
		     connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);
		     execute (update temp_obj_asoc set FOLDER_OBJECT_ASSOC_SK=FOLDER_OBJECT_ASSOC_SK_SEQ.nextval) by oracle; 	     		  	
		  	 execute (insert into DCM_FOLDER_OBJECT_ASSOC (FOLDER_SK,FOLDER_OBJECT_TYPE_CD,FOLDER_OBJECT_SK,FOLDER_OBJECT_PATH,FOLDER_OBJECT_ASSOC_SK)
			 select FOLDER_SK,FOLDER_OBJECT_TYPE_CD,FOLDER_OBJECT_SK,FOLDER_OBJECT_PATH,FOLDER_OBJECT_ASSOC_SK from temp_obj_asoc ) by oracle;
			 execute (commit) by oracle;
			 disconnect from oracle;
		  quit;
	 	%end;
		
		*Create temp ruleflow_hst table*;
		%if (%sysfunc(exist(brm.temp_ruleflow_hst)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_ruleflow_hst;
			quit;
		%end; 

		data brm.temp_ruleflow_hst (drop=RULE_FLOW_PATH);
			set brm.temp_ruleflow;
			ENVERS_HST_SK=&version_no;
			ENVERS_CHG_TYPE_CD=&no;
		run;

		
		*Create temp ruleflow_node and ruleflow_node_hst table *;	
		%if (%sysfunc(exist(brm.ruleflow_node)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.ruleflow_node;
			quit;
		%end; 
		proc sql noprint;			
			create table brm.ruleflow_node as
			select distinct
	               c.RULE_FLOW_SK,
	               a.NODE_TYPE_CD, 
				   a.NODE_OBJECT_SK, 
				   a.user                    as RULE_FLOW_NODE_CREATE_USER_NM,
				   a.user                    as RULE_FLOW_NODE_MOD_USER_NM,
				   a.RULE_SET_SECTION_CODE   as NODE_SECTION_CD,			 
				   a.sk                      as RULE_FLOW_NODE_SK,			  			   
				   b.ENVERS_HST_DTTM         as RULE_FLOW_NODE_CREATE_DTTM,
				   b.ENVERS_HST_DTTM         as RULE_FLOW_NODE_MOD_DTTM

			from   &data                     as a,			       
				   brm.DCM_ENVERS_INFO       as b,
				   brm.temp_ruleflow         as c

			where  b.ENVERS_HST_SK=&version_no and
				  (a.RULE_FLOW_SK=c.RULE_FLOW_SK or
	               missing(a.RULE_FLOW_SK));				 
		quit;

		proc sql noprint;
		  connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);
		  execute (update ruleflow_node set RULE_FLOW_NODE_SK=RULE_FLOW_NODE_SK_SEQ.nextval) by oracle;
		  execute(commit) by oracle;
		  disconnect from oracle;
		quit;

		%if (%sysfunc(exist(brm.ruleflow_node_hst)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.ruleflow_node_hst;
			quit;
		%end;

		data brm.ruleflow_node_hst;
			set brm.ruleflow_node;
			ENVERS_HST_SK=&version_no;
			ENVERS_CHG_TYPE_CD=0;
		run;

		%if (%sysfunc(exist(brm.ruleflow_node_hst_1)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.ruleflow_node_hst_1;
			quit;
		%end;

		proc sql noprint;
			create table brm.ruleflow_node_hst_1 as 
		    select a.ENVERS_CHG_TYPE_CD,
				   a.ENVERS_HST_SK,
				   a.RULE_FLOW_SK,
				   a.NODE_TYPE_CD,
				   a.NODE_OBJECT_SK,
	               a.RULE_FLOW_NODE_CREATE_USER_NM,
	               a.RULE_FLOW_NODE_MOD_USER_NM,
	               a.NODE_SECTION_CD,
	               a.RULE_FLOW_NODE_CREATE_DTTM,
	               a.RULE_FLOW_NODE_MOD_DTTM,
	               b.RULE_FLOW_NODE_SK
	       from    brm.ruleflow_node_hst         as a,
		           brm.DCM_RULE_FLOW_NODE        as b
		   where   a.RULE_FLOW_SK=b.RULE_FLOW_SK and
		   		   a.NODE_TYPE_CD=b.NODE_TYPE_CD and
				   a.NODE_OBJECT_SK=b.NODE_OBJECT_SK;
		quit;

		%brm_update_history_table(table=ruleflow_node_hst_1);

		*Create temp rule flow node link table and rule flow node link history table*;
		proc sql noprint;
			create table ruleflow_node as
			select b.RULE_FLOW_SK,
				   a.NODE_TYPE_CD,
			       a.order,			   
				   b.RULE_FLOW_NODE_SK

		   from    &data as a,
		           brm.ruleflow_node as b

		   where  (a.RULE_FLOW_SK=b.RULE_FLOW_SK  or
		           missing(a.RULE_FLOW_SK))       and
		   		   a.NODE_TYPE_CD=b.NODE_TYPE_CD  and
				   a.NODE_OBJECT_SK=b.NODE_OBJECT_SK

		   order by RULE_FLOW_NODE_SK,
		   			order;
		quit;

		proc sql noprint;
			create table node_link as select distinct
				RULE_FLOW_NODE_SK as FROM_RULE_FLOW_NODE_SK,
				(select rn2.RULE_FLOW_NODE_SK from ruleflow_node rn2 
						where rn2.rule_flow_sk = rn.rule_flow_sk and rn2.order = (rn.order + 1) ) as TO_RULE_FLOW_NODE_SK,
				"&brm_user" as RF_NODE_LINK_CREATE_USER_NM,
				. as RULE_FLOW_NODE_LINK_SK
			from 
				ruleflow_node rn
			where
				rn.order ne (select max(order) from ruleflow_node rni where rni.rule_flow_sk = rn.rule_flow_sk);
		quit;
		
		%if (%sysfunc(exist(brm.temp_node_link)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_node_link;
			quit;
		%end;

		proc sql noprint;
			create table brm.temp_node_link as
			select distinct
			       a.*,
				   b.ENVERS_HST_DTTM as RF_NODE_LINK_CREATE_DTTM

		    from   node_link as a,
				   brm.DCM_ENVERS_INFO       as b

			where  b.ENVERS_HST_SK=&version_no;
			       
		quit;

		*update RULE_FLOW_NODE_LINK_SK to RULE_FLOW_NODE_SK_SEQ.nextval*;
		proc sql noprint;
		  connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);
		  execute (update temp_node_link set RULE_FLOW_NODE_LINK_SK=RULE_FLOW_NODE_SK_SEQ.nextval) by oracle;
		  execute(commit) by oracle;
		  disconnect from oracle;
		quit;

		%if (%sysfunc(exist(brm.temp_node_link_hst)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_node_link_hst;
			quit;
		%end;

		data brm.temp_node_link_hst;
			set brm.temp_node_link;
			ENVERS_HST_SK=&version_no;
			ENVERS_CHG_TYPE_CD=0;
		run;

		%if (%sysfunc(exist(brm.temp_node_link_hst_1)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_node_link_hst_1;
			quit;
		%end;

		proc sql noprint;
			create table brm.temp_node_link_hst_1 as 
			select a.ENVERS_HST_SK,
				   a.ENVERS_CHG_TYPE_CD,
				   a.FROM_RULE_FLOW_NODE_SK,
				   a.TO_RULE_FLOW_NODE_SK,
				   a.RF_NODE_LINK_CREATE_USER_NM,
				   a.RF_NODE_LINK_CREATE_DTTM,
				   b.RULE_FLOW_NODE_LINK_SK

		   from    brm.temp_node_link_hst      as a,
		           brm.DCM_RULE_FLOW_NODE_LINK as b

		   where   a.FROM_RULE_FLOW_NODE_SK=b.FROM_RULE_FLOW_NODE_SK and
		           a.TO_RULE_FLOW_NODE_SK=b.TO_RULE_FLOW_NODE_SK;
		quit;

		%brm_update_history_table(table=temp_node_link_hst_1);
		
		%if (%sysfunc(exist(work.distinct_byterms)) eq 1) %then %do;
			proc sql noprint;
				drop table distinct_byterms;
			quit;
		%end;
		
		proc sql noprint;
			create table distinct_byterms as 
						select distinct BY_TERM, 
										VOCAB_NM, 
										RULE_FLOW_SK 
										from &data where not missing(BY_TERM);
		quit;

		*Create temp byterm table*;
		data byterm(keep=RULE_FLOW_SK VOCAB_NM BY_TERM count);
			set distinct_byterms;
			where not missing(BY_TERM);
			count=count(BY_TERM,',');			
		run;

		data byterm1(drop=count BY_TERM);
			length term $32;
			set byterm;
			where count=0;
			TERM_ORDER_NO=0;
			term=BY_TERM;
		run;

		data byterm2;
			set byterm;
			where count>0;
		run;

		%if %brm_obscount(byterm2)>0 %then %do;
			proc sql noprint;
				select count(*), max(count) into 
	                   :count,
					   :max
				from byterm2;
			quit;
			%let count=&count;
			%let max=&max;
		
			proc sql noprint;
				select BY_TERM,count,RULE_FLOW_SK,VOCAB_NM into 
	                   :term1  - :term&count,
					   :count1 - :count&count,
					   :sk1    - :sk&count,
					   :vocab1 - :vocab&count
				from byterm2;
			quit;	
			
			data byterm3;			
				length term term1 - term&max $32 VOCAB_NM $100;;
				%do i=1 %to &count;
					RULE_FLOW_SK=&&sk&i;
					VOCAB_NM="&&vocab&i";
					%do j=1 %to &&count&i;				    
						term&j=scan("&&term&i",&j,',');
						term=scan("&&term&i",-1,',');				
					%end;
					output;
				%end;			
			run;

			data byterm30(keep=RULE_FLOW_SK VOCAB_NM term TERM_ORDER_NO);
				set byterm3;
				TERM_ORDER_NO=&max;			
			run;
		    
			%do i=1 %to &max;
				data byterm3&i(keep=RULE_FLOW_SK VOCAB_NM term TERM_ORDER_NO);
					set byterm3(drop=term);
					rename term&i=term;
					%let term_no=%eval(&i-1);
					TERM_ORDER_NO=&term_no;	
				run;			
	            
				data byterm30;
					set byterm30 byterm3&i;				
				run;

				%if (%sysfunc(exist(byterm3&i)) eq 1) %then %do;
					proc sql noprint;
						drop table byterm3&i;
					quit;
				%end;
			%end;
		
			data temp_byterm;
				set byterm1 byterm30;
				GROUP_TERM_CREATE_USER_NM="&brm_user";
				GROUP_TERM_MOD_USER_NM="&brm_user";
				GROUP_TERM_SK=.;
			run;
		%end;
		%else %if %brm_obscount(byterm2)=0 %then %do;
			data temp_byterm;
				set byterm1;
				GROUP_TERM_CREATE_USER_NM="&brm_user";
				GROUP_TERM_MOD_USER_NM="&brm_user";
				GROUP_TERM_SK=.;
			run;
		%end;

		%if (%sysfunc(exist(brm.byterm)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.byterm;
			quit;
		%end;

		proc sql noprint;
			create table brm.byterm as
			select a.GROUP_TERM_SK,
				   a.TERM_ORDER_NO,
				   a.GROUP_TERM_CREATE_USER_NM,
				   a.GROUP_TERM_MOD_USER_NM,
				   a.RULE_FLOW_SK,
				   b.VOCAB_TERM_SK,
				   c.ENVERS_HST_DTTM as GROUP_TERM_CREATE_DTTM,
				   c.ENVERS_HST_DTTM as GROUP_TERM_MOD_DTTM

			from  temp_byterm               as a,
			      brm.DCM_VOCABULARY_TERM   as b,
				  brm.DCM_ENVERS_INFO       as c,
				  brm.DCM_VOCABULARY_ENTITY as d,
				  brm.DCM_VOCABULARY        as e

			where a.VOCAB_NM=e.VOCAB_NM                and
			      a.term=b.VOCAB_TERM_NM            and
				  b.VOCAB_ENTITY_SK=d.VOCAB_ENTITY_SK  and
				  d.VOCAB_SK=e.VOCAB_SK                and
				  c.ENVERS_HST_SK=&version_no;
		quit;

		*update GROUP_TERM_SK to GROUP_TERM_SK_SEQ.nextval*;
		proc sql noprint;
		  connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);
		  execute (update byterm set GROUP_TERM_SK=GROUP_TERM_SK_SEQ.nextval) by oracle;
		  execute(commit) by oracle;
		  disconnect from oracle;
		quit;

		%if (%sysfunc(exist(brm.byterm_hst)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.byterm_hst;
			quit;
		%end;

		data brm.byterm_hst;
			set brm.byterm;
			ENVERS_HST_SK=&version_no;
			ENVERS_CHG_TYPE_CD=0;
		run;

		%if (%sysfunc(exist(brm.byterm_hst_1)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.byterm_hst_1;
			quit;
		%end;

		proc sql noprint;
			create table brm.byterm_hst_1 as 
			select a.ENVERS_HST_SK,
				   a.ENVERS_CHG_TYPE_CD,
				   a.VOCAB_TERM_SK,
				   a.TERM_ORDER_NO,
				   a.GROUP_TERM_CREATE_USER_NM,
				   a.GROUP_TERM_CREATE_DTTM,
				   a.GROUP_TERM_MOD_USER_NM,
				   a.GROUP_TERM_MOD_DTTM,
				   a.RULE_FLOW_SK,
				   b.GROUP_TERM_SK

		   from    brm.byterm_hst      as a,
		           brm.DCM_GROUP_TERM  as b

		   where   a.RULE_FLOW_SK=b.RULE_FLOW_SK and
		           a.VOCAB_TERM_SK=b.VOCAB_TERM_SK;
		quit;

		%brm_update_history_table(table=byterm_hst_1);
		
		*Insert temp_table records into DBMS tables*;
		%macro brm_update_table(table=,temp_table=);
			*Store table column to macro variable*;					
			proc contents data=brm.&temp_table out=&temp_table noprint;
			run;

			proc sql noprint;
				select name into :&temp_table.1 separated by ','
				from &temp_table;
			quit;		

			proc datasets lib=work memtype=data noprint;
			   %if (%sysfunc(exist(&temp_table)) eq 1) %then %do;
			      DELETE &temp_table;
			   %end;
			run;
			quit;		
		
			*Insert temp table records into DBMS corresponding table*;		
			filename import temp;
			data _null_;
				file import LRECL=32767;								
				put 'proc sql noprint;';
				put '	connect to oracle as oracle (path=&brm_dbms_path authdomain=&brm_authdomain);';	
				put	"   execute (insert into &table (&&&temp_table.1)";
			    put "   select &&&temp_table.1 from &temp_table) by oracle;";				
				put '   execute(commit) by oracle;';	
	            put '   execute(drop table &temp_table) by oracle;';	
				put '   disconnect from oracle;';	
				put 'quit;';
			run;
			
			%include import;

			%if (%sysfunc(fileref(import)) eq 0) %then %do;
		      filename import clear;
		    %end;
		%mend brm_update_table;
		*Add delete rcords to history tables*;
		%brm_update_table(table=%str(DCM_RULE_FLOW_NODE_HST),temp_table=%str(ruleflow_node_hst_1));
		%brm_update_table(table=%str(DCM_RULE_FLOW_NODE_LINK_HST),temp_table=%str(temp_node_link_hst_1));
		%brm_update_table(table=%str(DCM_GROUP_TERM_HST),temp_table=%str(byterm_hst_1));
		
		*Delete records from DBMS tables*;
		%if "&update"="Y" %then %do;
			proc sql noprint;
				delete * from brm.DCM_RULE_FLOW_NODE_LINK
				where  FROM_RULE_FLOW_NODE_SK in (select RULE_FLOW_NODE_SK  from brm.DCM_RULE_FLOW_NODE where RULE_FLOW_SK in 
									             (select RULE_FLOW_SK from update_ruleflow )) or 
					   TO_RULE_FLOW_NODE_SK in (select RULE_FLOW_NODE_SK  from brm.DCM_RULE_FLOW_NODE where RULE_FLOW_SK in 
									             (select RULE_FLOW_SK from update_ruleflow ))	
								  ;

				delete * from brm.DCM_GROUP_TERM
				where  RULE_FLOW_SK in (select RULE_FLOW_SK from update_ruleflow);

				delete * from brm.DCM_RULE_FLOW_OUT_NODE_X_TERM 
				where RULE_FLOW_NODE_SK in (select RULE_FLOW_NODE_SK from update_ruleflow ur, brm.DCM_RULE_FLOW_NODE rn
											where rn.rule_flow_sk = ur.rule_flow_sk);

				delete * from brm.DCM_RULE_FLOW_NODE 
				where RULE_FLOW_SK in (select RULE_FLOW_SK from update_ruleflow);					
		    quit;	
		%end;

		%if (%sysfunc(exist(brm.temp_ruleflow2)) eq 1) %then %do;
			proc sql noprint;
				drop table brm.temp_ruleflow2;
			quit;
		%end; 
        /*temp_ruleflow actually has an extra column we need to get rid of
		  because when you do updates you don't have an SK to link to. */	
		data brm.temp_ruleflow2 (drop=RULE_FLOW_PATH);
			set brm.temp_ruleflow;
		run;

		*Insert temp table records to DBMS tables*;
		%if "&update" eq "N" %then %do;
			%brm_update_table(table=%str(DCM_RULE_FLOW),temp_table=%str(temp_ruleflow2));
		%end;
		%brm_update_table(table=%str(DCM_RULE_FLOW_HST),temp_table=%str(temp_ruleflow_hst));
		%brm_update_table(table=%str(DCM_RULE_FLOW_NODE),temp_table=%str(ruleflow_node));
		%brm_update_table(table=%str(DCM_RULE_FLOW_NODE_HST),temp_table=%str(ruleflow_node_hst));
		%brm_update_table(table=%str(DCM_RULE_FLOW_NODE_LINK),temp_table=%str(temp_node_link));
		%brm_update_table(table=%str(DCM_RULE_FLOW_NODE_LINK_HST),temp_table=%str(temp_node_link_hst));
		%brm_update_table(table=%str(DCM_GROUP_TERM),temp_table=%str(byterm));
		%brm_update_table(table=%str(DCM_GROUP_TERM_HST),temp_table=%str(byterm_hst));

	%mend brm_import_table;

	*Verify whether create_ruleflow has records *;
	%if %brm_obscount(create_ruleflow) =0 %then %do;
		%put NOTE: There are no ruleflows to create;		
	%end;
	%else %if %brm_obscount(create_ruleflow) >0 %then %do;
		%brm_import_table(update=N);
	%end;	

	*Verify whether update_ruleflow has records*;
	%if %brm_obscount(update_ruleflow) =0 %then %do;
		%put NOTE: There are no ruleflow to update;		
	%end;
	%else %if %brm_obscount(update_ruleflow) >0 %then %do;
		%brm_import_table(update=Y);
	%end;

	%exit:
	*Store rejected records in reject dataset*;
	%if %brm_obscount(ruleflow)>0 %then %do;
		%let condition=%str(RULE_FLOW_NM not in (select RULE_FLOW_NM from ruleflow ));
	%end;	
	%else %do;
		%let condition=1;
	%end;

	proc sql noprint;
		create table reject as 
		select * 
		from   csv
		where  &condition;
	quit;

	*Export reject dataset to csv*;
	proc export data=work.reject
	  outfile="&reject"
	  dbms=csv 
	  replace;
	run;

    * Remove lock file ;
	%if (%sysfunc(exist(brm.lock_import_rule_flow)) eq 1) %then %do;
	  	 proc sql noprint;
			drop table brm.lock_import_rule_flow;
		 quit;
	%end;

    %lockexit:  * Keep the lock intact ;

	* Delete temp table in work library *;	
	 proc datasets lib=work memtype=data noprint;
	   %if (%sysfunc(exist(work.csv)) eq 1) %then %do;
	      DELETE csv;
	   %end;

	   %if (%sysfunc(exist(work.input)) eq 1) %then %do;
	      DELETE input;
	   %end;

	   %if (%sysfunc(exist(work.ruleflow_rsd)) eq 1) %then %do;
	      DELETE ruleflow_rsd;
	   %end;

	   %if (%sysfunc(exist(work.ruleflow_rsv)) eq 1) %then %do;
	      DELETE ruleflow_rsv;
	   %end;

	   %if (%sysfunc(exist(work.ruleflow_deletes)) eq 1) %then %do;
	      DELETE ruleflow_deletes;
	   %end;
    
	   %if (%sysfunc(exist(work.fullpathfolderdata)) eq 1) %then %do;
	      DELETE fullpathfolderdata;
	   %end;

	   %if (%sysfunc(exist(work.missingrulesets)) eq 1) %then %do;
	      DELETE missingrulesets;
	   %end;

	   %if (%sysfunc(exist(work.ruleflow)) eq 1) %then %do;
	      DELETE ruleflow;
	   %end;

	   %if (%sysfunc(exist(work.ruleflow_node)) eq 1) %then %do;
	      DELETE ruleflow_node;
	   %end;

	   %if (%sysfunc(exist(work.from_node)) eq 1) %then %do;
	      DELETE from_node;
	   %end;

	   %if (%sysfunc(exist(work.to_node)) eq 1) %then %do;
		      DELETE to_node;
	   %end;

	   %if (%sysfunc(exist(work.update_ruleflow)) eq 1) %then %do;
	      DELETE update_ruleflow;
	   %end;

	   %if (%sysfunc(exist(work.temp_obj_asoc)) eq 1) %then %do;
	      DELETE temp_obj_asoc;
	   %end;

	   %if (%sysfunc(exist(work.temp_obj_asoc_1)) eq 1) %then %do;
	      DELETE temp_obj_asoc_1;
	   %end;

	   %if (%sysfunc(exist(work.temp_obj_asoc_2)) eq 1) %then %do;
	      DELETE temp_obj_asoc_2;
	   %end;

	   %if (%sysfunc(exist(work.distinct_byterms)) eq 1) %then %do;
	      DELETE distinct_byterms;
	   %end;

	   %if (%sysfunc(exist(work.byterm)) eq 1) %then %do;
	      DELETE byterm;
	   %end;

	   %if (%sysfunc(exist(work.byterm1)) eq 1) %then %do;
	      DELETE byterm1;
	   %end;

	   %if (%sysfunc(exist(work.byterm2)) eq 1) %then %do;
	      DELETE byterm2;
	   %end;

	   %if (%sysfunc(exist(work.byterm3)) eq 1) %then %do;
	      DELETE byterm3;
	   %end;

	   %if (%sysfunc(exist(work.byterm30)) eq 1) %then %do;
	      DELETE byterm30;
	   %end;

	   %if (%sysfunc(exist(work.Create_ruleflow)) eq 1) %then %do;
	      DELETE Create_ruleflow;
	   %end;

	   %if (%sysfunc(exist(work.folderdata)) eq 1) %then %do;
	      DELETE folderdata;
	   %end;

	   %if (%sysfunc(exist(work.end)) eq 1) %then %do;
	      DELETE end;
	   %end;

	   %if (%sysfunc(exist(work.Final)) eq 1) %then %do;
	      DELETE Final;
	   %end;

	   %if (%sysfunc(exist(work.Init)) eq 1) %then %do;
	      DELETE Init;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_end)) eq 1) %then %do;
	      DELETE Invalid_end;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_init)) eq 1) %then %do;
	      DELETE Invalid_init;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_main)) eq 1) %then %do;
	      DELETE Invalid_main;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_order)) eq 1) %then %do;
	      DELETE Invalid_order;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_section)) eq 1) %then %do;
	      DELETE Invalid_section;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_start)) eq 1) %then %do;
	      DELETE Invalid_start;
	   %end;

	   %if (%sysfunc(exist(work.Invalid_term)) eq 1) %then %do;
	      DELETE Invalid_term;
	   %end;

	   %if (%sysfunc(exist(work.Main)) eq 1) %then %do;
	      DELETE Main;
	   %end;

	   %if (%sysfunc(exist(work.Node_link)) eq 1) %then %do;
	      DELETE Node_link;
	   %end;

	   %if (%sysfunc(exist(work.order)) eq 1) %then %do;
	      DELETE order;
	   %end;

	   %if (%sysfunc(exist(work.section)) eq 1) %then %do;
	      DELETE section;
	   %end;

	   %if (%sysfunc(exist(work.start)) eq 1) %then %do;
	      DELETE start;
	   %end;

	   %if (%sysfunc(exist(work.Temp_byterm)) eq 1) %then %do;
	      DELETE Temp_byterm;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_1)) eq 1) %then %do;
	      DELETE Ruleflow_1;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_2)) eq 1) %then %do;
	      DELETE Ruleflow_2;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_3)) eq 1) %then %do;
	      DELETE Ruleflow_3;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_4)) eq 1) %then %do;
	      DELETE Ruleflow_4;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_5)) eq 1) %then %do;
	      DELETE Ruleflow_5;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_6)) eq 1) %then %do;
	      DELETE Ruleflow_6;
	   %end;

	   %if (%sysfunc(exist(work.Ruleflow_7)) eq 1) %then %do;
	      DELETE Ruleflow_7;
	   %end;

	   %if (%sysfunc(exist(work.reject)) eq 1) %then %do;
	      DELETE reject;
	   %end;
	run;
	quit;

	%if (%sysfunc(exist(brm.temp_obj_asoc)) eq 1) %then %do;
		proc sql noprint;
			drop table brm.temp_obj_asoc;
		quit;
	%end; 

	%if (%sysfunc(exist(brm.temp_ruleflow)) eq 1) %then %do;
		proc sql noprint;
			drop table brm.temp_ruleflow;
		quit;
	%end; 

	*deallocate BRM libname*;
    %brm_release(); 

%mend brm_import_rule_flow;
