
/******************************************************************************/
/*  Copyright (c) 2005-2012 by SAS Institute, Inc., Cary, NC, USA.            */
/*  All rights reserved.                                                      */
/******************************************************************************/
/*  NAME:        brs_decision_flow.sas                                        */
/*  DESCRIPTION: Generate SAS code from Business Rule Flow deployed xml file. */
/*  PARAMETERS:                                                               */
/*   InputTable   input data set (unmapped)                                   */
/*   Mapping      maps inputTable columns to rule term names                  */
/*                ** must be supplied unless logicCodeLoc= is specified       */
/*   Filelocation location of the rule flow xml file,                         */
/*   DebugCodeLoc location of the generated debug code (full code capture)    */
/*   LogicCodeLoc location of where to write the captured logic code (only)   */
/*   dbms_path    path= for db libname statement (tokenized for config)       */
/*   authdomain   dbms authentication domain (tokenized for config)           */
/*   RuleSets     Common Delimited List of Rule Set SK's to generate code for */
/*                if the value is empty ALL rule sets will be generated       */
/*   RuleFire     whether rule fire table is populated or not                 */
/*                   Y means rule fire table is populated,                    */
/*                   N means rule fire table is not populated. (default)      */
/*                                                                            */
/*  If logicCodeLoc= is specified, then the logic code only is written there, */
/*  no other code is generated, and no other code is run.                     */
/*                                                                            */
/*  logicCodeLoc takes precedence over DebugCodeLoc and the normal invocation */
/*                                                                            */
/*  RuleFire is always honored in code generation and execution.              */
/*                                                                            */
/*  If the user has already assigned a libref of BRM, it is concatenated in   */
/*  front of the database accessed via the db allocation parameters           */
/*                                                                            */
/*  PRODUCT:     Business Rule Management                                     */
/*  OWNER:       Jane Xue, Carl Sommer                                        */
/*  NOTES:       This macro is used to parse xml generated from decision flow,*/
/*               this code is tested against a standard format and should     */   
/*               NOT be modified by users.                                    */
/*                                                                            */
/*    Aug 30, 2013  - add special flag / parm to skip and use code that       */
/*                    was previously generated from a prior run               */
/******************************************************************************/ 

/* first define several macros that are useful in defining rules */
/* each of these returns 1 if found or 0 if not found.           */
/*--- match any in a list (list can be numeric or character)---*/
%macro brm_MatchAny(term,
     o1,o2,o3,o4,o5,o6,o7,o8,o9,o10,
     o11,o12,o13,o14,o15,o16,o17,o18,o19,o20,
     o21,o22,o23,o24,o25,o26,o27,o28,o29,o30,
     o31,o32,o33,o34,o35,o36,o37,o38,o39,o40,
     o41,o42,o43,o44,o45,o46,o47,o48,o49,o50); 
     
  &term and
  ( 
  %do i= 1 %to 50 ;
     %if "&&o&i" ne "" %then %do;
        %if &i gt 1 %then %do ;
   or
        %end ;
   (&term eq &&o&i)
     %end ;
  %end ; 
  ) 
%mend brm_MatchAny;

/*--- does the term begin with the (sub) string ---*/
%macro brm_StartsWith(term,
     o1,o2,o3,o4,o5,o6,o7,o8,o9,o10,
     o11,o12,o13,o14,o15,o16,o17,o18,o19,o20,
     o21,o22,o23,o24,o25,o26,o27,o28,o29,o30,
     o31,o32,o33,o34,o35,o36,o37,o38,o39,o40,
     o41,o42,o43,o44,o45,o46,o47,o48,o49,o50); 
     
  &term 

  %do i= 1 %to 50 ;
     %if "&&o&i" ne "" %then %do;
        %if &i eq 1 %then %do ; 
   and  ( 
		%end;
        %if &i gt 1 %then %do ;
   or
        %end ;
   (ksubstr(&term,1,min(ifn(klength(&term),klength(&term),1),
                        ifn(klength(&&o&i),klength(&&o&i),1))) eq &&o&i )
     %end ;
  %end ; 
  ) 
%mend brm_StartsWith;


/*--- does the term contain one of these (sub) strings ---*/
%macro brm_Contains(term,
     o1,o2,o3,o4,o5,o6,o7,o8,o9,o10,
     o11,o12,o13,o14,o15,o16,o17,o18,o19,o20,
     o21,o22,o23,o24,o25,o26,o27,o28,o29,o30,
     o31,o32,o33,o34,o35,o36,o37,o38,o39,o40,
     o41,o42,o43,o44,o45,o46,o47,o48,o49,o50); 
     
  &term and
  ( 
  %do i= 1 %to 50 ;
     %if "&&o&i" ne "" %then %do;
        %if &i gt 1 %then %do ;
   or
        %end ;
   (index(&term,&&o&i))
     %end ;
  %end ; 
  ) 
%mend brm_Contains;

/*--- does the term end with the (sub) string ---*/
%macro brm_EndsWith(term,
     o1,o2,o3,o4,o5,o6,o7,o8,o9,o10,
     o11,o12,o13,o14,o15,o16,o17,o18,o19,o20,
     o21,o22,o23,o24,o25,o26,o27,o28,o29,o30,
     o31,o32,o33,o34,o35,o36,o37,o38,o39,o40,
     o41,o42,o43,o44,o45,o46,o47,o48,o49,o50); 
     
  &term 

  %do i= 1 %to 50 ;
     %if "&&o&i" ne "" %then %do;
        %if &i eq 1 %then %do ; 
   and  ( 
		%end;
        %if &i gt 1 %then %do ;
   or
        %end ;
 
   (ksubstr(kleft(reverse(&term)),1,min(ifn(klength(&term),klength(&term),1),
                        ifn(klength(&&o&i),klength(&&o&i),1))) eq reverse(&&o&i) )
     %end ;
  %end ; 
  ) 
%mend brm_EndsWith;

/*--- does the term with a regex pattern  ---*/
%macro brm_prxmatch(term,
     o1,o2,o3,o4,o5,o6,o7,o8,o9,o10,
     o11,o12,o13,o14,o15,o16,o17,o18,o19,o20,
     o21,o22,o23,o24,o25,o26,o27,o28,o29,o30,
     o31,o32,o33,o34,o35,o36,o37,o38,o39,o40,
     o41,o42,o43,o44,o45,o46,o47,o48,o49,o50); 
     
  &term and
  ( 
  %do i= 1 %to 50 ;
     %if "&&o&i" ne "" %then %do;
        %if &i gt 1 %then %do ;
   or
        %end ;
   (prxmatch((&&o&i),(&term)))
     %end ;
  %end ; 
  ) 
%mend brm_prxmatch;

%macro brs_decision_flow(InputTable=, 
                         Mapping=, 
                         Filelocation=,
                         DebugCodeLoc=%str(),
                         logicCodeLoc=,
                         dbms_path=,
                         authdomain=,
                         RuleSets=,
                         RuleFire=N,
                         useExistingCodeGen=N);

   * save options settings ;
   %local _opt_LRECL  _opt_S  _opt_S2 _opt_S2V;
   %let _opt_LRECL  = %sysfunc(getoption(lrecl)) ;
   %let _opt_S      = %sysfunc(getoption(s)) ;
   %let _opt_S2     = %sysfunc(getoption(s2)) ;
   %let _opt_S2V     = %sysfunc(getoption(s2v)) ; 


   * override settings ;
   options LRECL=MAX S=MAX S2=MAX S2V=MIN;

    * Include some other useful macros we need ;
    %include "@sasmacro.dir@brm_alloc.sas";
    %include "@sasmacro.dir@brm_release.sas";
    %include "@sasmacro.dir@brm_obscount.sas";
                           
    * fasthpath to execution is to use previously-generated code if it exists ;
	%if %symexist(BRM_use_existing_code) eq 1 %then %do;
	   %let useExistingCodegen = %upcase(&BRM_use_existing_code);
	%end;

    %if ((%upcase(&useExistingCodegen) eq Y)      and
         ("&DebugCodeLoc."  ne "")) %then %do;
         %* SAS does not short-circuit, so break apart into two statements ;
		 %if (%sysfunc(fileexist(&debugcodeloc)) eq 1) %then %do;
           filename genfile "&DebugCodeLoc." LRECL=32767 MOD encoding='utf-8';  
          %goto fastPath;
		%end;
    %end; 
                       
%normal:
    * InputTable=, Mapping= must exist if codeGen is not specified ;
    %if ("&logicCodeLoc." ne "") %then %do;
       filename genfile "&logicCodeLoc." LRECL=32767 encoding='utf-8';
    %end;
    %else %do;
       * InputTable= must exist in this context ;
       %if ("&InputTable." eq "") %then %do;
          %put "ERROR: No input table specified";
          %goto exit;
       %end;
       %else %if (%sysfunc(exist(&InputTable)) eq 0) AND
                 (%sysfunc(exist(&InputTable,VIEW)) eq 0) %then %do;
          %put ERROR: Input table &inputTable does not exist.;
          %goto exit;
       %end;

       * Mapping= is required in this context;
       %if ("&Mapping." eq "") %then %do;
           %put ERROR: Mapping= not specified but required for this invocation.;
           %goto exit;
       %end;
       %else %if (%sysfunc(exist(&Mapping)) eq 0) %then %do;
          %put ERROR: Mapping table &mapping does not exist.;
          %goto exit;
       %end;

       %if ("&DebugCodeLoc."  ne "") %then %do;
         filename genfile "&DebugCodeLoc." LRECL=32767 encoding='utf-8';
       %end;
       %else %do;
         filename genfile temp LRECL=32767 encoding='utf-8';
       %end;

       * Validate that code location fileref is assigned ;
       %if (%sysfunc(fileref(genfile)) gt 0) %then %do;
          %put ERROR: Code output location could not be assigned.;
          %goto exit;
       %end;
    %end;

    %let RULEFIREDSETID = 1;
    %let EXECUTIONSETID = 2;
    %let PARAMETERSETID = 3;

    libname &Filelocation. xmlv2 xmlmap="@sasmisc.dir@/DecisionFlow-v4.map";

    * xml library data cannot be read directly, need to copy the data to ;
    * another library,copy the data sets to work. ;
    proc copy in=&filelocation. out=work;
    run;               
    
    * Compress duplicate records in Lookup table ;
    proc sql noprint;
        create table lookups as 
        select distinct *
        from lookup;  
    quit;

    data _null_;
       file genfile LRECL=32767;
       put "/* Code generated on &SYSDATE at &SYSTIME by &SYSUSERID */";
       %if ("&logicCodeLoc" eq "") %then %do;
          * Set NOQUOTELENMAX option to fix S0872475 ;
          put "options NOQUOTELENMAX;";
       %end;
       put;
    run;

    * If the ruleset has lookup, then generate database libref ;
    proc sql noprint;
       select count(*) into :LCount
       from lookups;          
    quit;

    * Generate the libname statement for the included code (not logic-only) ;
    %if ((&LCount. gt 0) and ("&logicCodeLoc" eq "")) %then %do;             
       %brm_alloc(dbms=ORACLE,dbms_path=&dbms_path,authdomain=&authdomain,access=readonly);
       data _null_;
         file genfile mod LRECL=32767;
         * macros we will want at runtime for fastpath ;
         put '%include "@sasmacro.dir@brm_alloc.sas";';
         put '%include "@sasmacro.dir@brm_release.sas";';
         put '%include "@sasmacro.dir@brm_obscount.sas";';
         put '%brm_alloc(dbms=ORACLE,';
         put "           dbms_path=&dbms_path,";
         put "           authdomain=&authdomain,access=readonly);";
       run;
    %end;

    * Put deploymentkey and folder key to macro variable ;
    %let DeploySk=;
    %let RuleFlowSk=;
    %let DeployName=;
    %let RuleFlowName=;

    proc sql noprint;
       select DeploymentSk, ruleFlowSk, deploymentName,ruleFlowName
       into :DeploySk, :RuleFlowSk, :DeployName, :RuleFlowName
       from decisionflow;
    quit;

    %let DeploySk=&DeploySk.;
    %let RuleFlowSk=&RuleFlowSk.;
    %let DeployName=&DeployName.;
    %let RuleFlowName=&RuleFlowName.;

    proc sort data=ruleaction;
        by rulesetid ruleid;
    run;

    %let NodeNo=;

    * Begin at analysis node ;
    proc sql noprint;
       select count(*) into :NodeNo
       from node
       where nodetype ne " ";
    quit;
    
    %let NodeNo=&NodeNo.;

    %if &NodeNo. le 0 %then %do;
      * No nodes to process.  Put out a message and end;
      %put "Decision flow does not contain valid node";
      %goto exit;
    %end;

    proc sql noprint;
       select startnodeid into :StartNodeID
       from nodes;
    quit;

    %let StartNodeID=&StartNodeID.;
     
   
    * Data step construction setup code - skip if only capturing logic ;
    %if ("&logicCodeLoc" eq "") %then %do;
       * Count output table number from &mapping table ;
       proc sql noprint;
          create table outtable as 
          select distinct datasetid
          from &Mapping.
          where upcase(type)="OUTPUT";
       quit;

       %let OutNo=;
       proc sql noprint;
          select count(*) into :OutNo
          from outtable;
       quit;   

       %let OutNo=&OutNo.;

       * Generate a keep statement for the output parameters which need to be maintained in the final output. ;
       proc sql noprint;
          %if &OutNo. < 2 %then %do;
             select datasetid into :OUTMAP1
             from outtable;        
          %end;
          %else %do;
             select datasetid into :OUTMAP1 - :OUTMAP&OutNo.
             from outtable;        
          %end;
       quit;       
         
       %do i=1 %to &OutNo.;
          %let SETID = &&OUTMAP&i;        
          %let OUTSTATE&i=&&OUTPUT&SETID;
       %end;
    %end; * full code capture ;

    * %brs_ruleset macro is used to handle ruleset node, we assume one node only has one ruleset ;
    %macro brs_ruleset(RuleSetID=, NextNodeID=, NodeID=);        
       proc sql noprint;
         create table rule_temp as 
           select a.cexpression,                   
                  a.rulesetid,
                  a.ruleid, 
                  b.aexpression,
				  b.actionid
           from rulecondition as a left join ruleaction as b
           on a.rulesetid=b.rulesetid and a.ruleid=b.ruleid ; 


          create table rules as
            select distinct rule.*, 
                   rule_temp.cexpression, 
                   rule_temp.aexpression,
				   rule_temp.actionid,
                   ruleset.rulesetname
            from rule, rule_temp,ruleset
            where rule.rulesetid = &RuleSetID.                          and 
                  rule.rulesetid=rule_temp.rulesetid=ruleset.rulesetid  and                  
                 rule.ruleid=rule_temp.ruleid
            order by rule.ExecutionSeqNbr,
                     rule_temp.actionID;

            %if (&StartNodeID eq &NodeID) %then %do;
                create table orderedrulesets as 
                  select rulesetid, rulesetname, 1 as order_no 
                  from ruleset 
                  where rulesetid = &RuleSetID.;
            %end;
            %else %do;
                select max(order_no) + 1 into :UPDATEDNUMBER 
                from orderedrulesets;
                insert into orderedrulesets 
                  select rulesetid, rulesetname, &UPDATEDNUMBER as order_no 
                  from ruleset 
                  where rulesetid = &RuleSetID.;
            %end;
       quit;  
         
       %if (("&RuleSets." eq "") or 
            (%index(&RuleSets.,&RuleSetID.) ne 0)) %then %do; 
          data _null_;
            file genfile mod LRECL=32767;
            put "*RuleSet &RuleSetID-- Starts Here*;";
          run;  
 
          data _null_ ;
             set rules ;
             by executionSeqNbr actionID;
             file genfile mod LRECL=32767;

             length conditionOperator $10;

			 if (first.executionSeqNbr and first.actionID) then do;
			   put;
               if (upcase(ruleJoinOperator) in ('ELSEIF','OR')) then do;
                  conditionOperator = 'else if ';         
               end;
               else do;
                  conditionOperator = 'if ' ;
               end;
               put conditionOperator cexpression ;
               put "then do ; ";
			 end;

             put aexpression ';';

             if (last.executionSeqNbr and last.actionID) then do;
               %if (%upcase("&RuleFire.") eq "Y" and ("&logicCodeLoc." eq "")) %then %do;
                  put "    RULE_FIRE_DTTM=datetime();";
                  put "    RULE_ACTION_FIRE_ID=uuidgen();";
                  put "    DEPLMT_SK=&DeploySk.;";
                  put "    DEPLMT_NM='&DeployName.';";
                  put "    RULE_FLOW_SK=&RuleFlowSk.;";
                  put "    RULE_FLOW_NM='&RuleFlowName.';";
                  put "    DEPLMT_EXECUTION_ID=symget('DEPLMT_EXECUTION_ID');";
                  %if %symexist(brmPrimaryEntityKey) %then %do;
                     %if NOT ( %sysevalf(%superq(brmPrimaryEntityKey)=,boolean)) %then %do;
                        put "    ENTITY_PRIMARY_KEY=&brmPrimaryEntityKey.;";    
                     %end;
                  %end;
                  %else %do;
                     put "    call missing(ENTITY_PRIMARY_KEY);";        
                  %end;  
                  %if %symexist(brmTransactionDttm)  %then %do;
                     %if NOT ( %sysevalf(%superq(brmTransactionDttm)=,boolean) ) %then %do;
                        put "    TRANSACTION_DTTM=&brmTransactionDttm.;";       
                     %end;
                  %end;
                  %else %do;
                     put "    call missing(TRANSACTION_DTTM);";      
                  %end;                   
                  put '    RULE_SET_SK=' rulesetid ';';
                  put '    RULE_SK=' ruleid ';';
                  put '    RULE_SET_NM="' RuleSetName '";';
                  put '    RULE_NM="' RuleName '";';
                  put "    output &&OUTSTATE&RULEFIREDSETID. ;";
               %end;
               put "end;";
               put;   
             end; 
          run;

          data _null_;
             file genfile mod LRECL=32767;
             put "*RuleSet &RuleSetID -- End Here*;";  
          run;
       %end;
       %brs_node(NodeID=&NextNodeID.);
   
    %mend brs_ruleset;

    %macro brs_complete(NodeId=);     
       %if ("&logicCodeLoc" eq "") %then %do;
          proc sql noprint;
             select dataSetOutputId
             into  :outputdatasetid
             from decisionComplete
             where nodeid=&NodeID.;
          quit;
          %let outputdatasetid=&outputdatasetid.;             
             
          * We only need to output the dataset referenced as the output dataset in the completion node ;
          data _null_;
             file genfile mod LRECL=32767;

             %if ((%upcase("&currentSection") eq "INIT") or
                  (%upcase("&currentSection") eq "")) %then %do;
                put "end;  /* close off _N_ eq 1 */";
             %end;
             %else %if (%upcase("&currentSection") eq "GROUPSTART") %then %do;
                put "end;  /* close off first.&finalByTerm */";
             %end;
             %else %if (%upcase("&currentSection") eq "GROUPEND") %then %do;
                put "end;  /* close off last.&finalByTerm */";
             %end;
             %else %if (%upcase("&currentSection") eq "FINAL") %then %do;
                put "end;  /* close off end=_last */";
             %end;
             put;

           
             %if (&byTermCount gt 0) %then %do;    
                put "if (last.&finalByTerm) then do;";
             %end;
             put "output &&OUTPUT&outputdatasetid. ;";
             %if (&byTermCount gt 0) %then %do;    
                put "end;";
             %end;
             %if (%symexist(LCount) and (&LCount. > 0)) %then %do;
                put ' if _last then do;';
                put '   lookupValues.delete();';  
                put ' end;';
             %end;         
          run;  
       %end; %* Complete code ;
    %mend brs_complete;
    
    %macro brs_node(NodeID=);   
       proc sql noprint;
          select nodetype, section into :NodeType, :section
          from node
          where nodeid=&NodeID.;
       quit;

       %let NodeType = &NodeType;
       %let Section = &Section;

       %if (%upcase("&NodeType.") eq "COMPLETE") %then %do;     
          * Reset node type to avoid infinite loops. ;
          %let NodeType=;         
          %brs_complete(NodeId=&NodeID);
          %goto exit; 
       %end;   
         
       %if (%upcase("&NodeType.") eq "RULESET") %then %do;
          * Reset node type to avoid infinite loops. ;
          %let NodeType=;
          proc sql noprint;
             select rulesetid, nextnodeid
             into  :RuleSetID, :NextNodeID
             from rulesetnode
             where nodeid=&NodeID.;
          quit;
            
          %let RuleSetID=&RuleSetID.;
          %let NextNodeID=&NextNodeID.;

          %* Determine if we need to end / start a new section of processing ;
          %if (%upcase("&currentSection") ne %upcase("&section")) %then %do; 
            data _null_ ;
              file genfile mod LRECL=32767;

              %if (%upcase("&section") eq "INIT") %then %do;
                 %if  (%brm_obscount(work.initRetains)) %then %do;
                    set work.initRetains (in=_init);
                    put 'retain ' termid +(-1) ';';
                %end;
              %end;

              %else %if(%upcase("&section") eq "GROUPSTART") %then %do;
                 %if ((%upcase("&currentSection") eq "INIT") or
                      (%upcase("&currentSection") eq "")) %then %do;
                    %if (%brm_obscount(work.groupStartRetains)) %then %do;
                       set work.groupStartRetains;
                    %end;
                    if _n_ eq 1 then do;
                       put "end;  /* close off _N_ eq 1 */";
                       put;
                       put "if (first.&finalByTerm) then do;";
                    end;
                 %end;
                 %if (%brm_obscount(work.groupStartRetains)) %then %do;
                    put 'retain ' termid +(-1) ';';
                 %end;
              %end;
              %else %if(%upcase("&section") eq "MAIN") %then %do;
                 %if ((%upcase("&currentSection") eq "INIT") or
                      (%upcase("&currentSection") eq "")) %then %do;
                    put "end;  /* close off _N_ eq 1 */";
                 %end;
                 %else %if (%upcase("&currentSection") eq "GROUPSTART") %then %do;
                    put "end;  /* close off first.&finalByTerm */";
                 %end;
                 put;
              %end;
              %else %if(%upcase("&section") eq "GROUPEND") %then %do;
                 %if ((%upcase("&currentSection") eq "INIT") or
                      (%upcase("&currentSection") eq "")) %then %do;
                    put "end;  /* close off _N_ eq 1 */";
                 %end;
                 %else %if (%upcase("&currentSection") eq "GROUPSTART") %then %do;
                    put "end;  /* close off first.&finalByTerm */";
                 %end;
                 put;
                 put "if (last.&finalByTerm) then do;";
              %end;
              %else %if(%upcase("&section") eq "FINAL") %then %do;
                 %if ((%upcase("&currentSection") eq "INIT") or
                      (%upcase("&currentSection") eq "")) %then %do;
                    put "end;  /* close off _N_ eq 1 */";
                 %end;
                 %else %if (%upcase("&currentSection") eq "GROUPSTART") %then %do;
                    put "end;  /* close off first.&finalByTerm */";
                 %end;
                 %else %if (%upcase("&currentSection") eq "GROUPEND") %then %do;
                    put "end;  /* close off last.&finalByTerm */";
                 %end;
                 put;
                 put "if (_last) then do;";
              %end;
            run;
            %let currentSection = &section;
          %end;
          %brs_ruleset(RuleSetID=&RuleSetID.,NextNodeID=&NextNodeID.,NodeID=&NodeID.);
       %end;

       %if (%upcase("&NodeType.") eq "SASCODE") %then %do;
          * Reset node type to avoid infinite loops. ;
          %let NodeType=;
          %brs_sascode;
       %end;
    %exit:
    %mend brs_node;

     
    * If we have are building all of the code, then store the execution ,  ;
    * parameter,rulefire, and output column properties to macro variable.  ;
    %if ("&logicCodeLoc" eq "") %then %do;
       proc sql noprint;
          select count(*) into :ExeNo
          from &mapping.
          where datasetid="&EXECUTIONSETID.";

          select count(*) into :ParNo
          from &mapping.
          where datasetid="&PARAMETERSETID.";

          select count(*) into :FireNo
          from &mapping.
          where datasetid="&RULEFIREDSETID.";

          select count(*) into :OutsNo
          from &mapping.
          where upcase(type)="OUTPUT"      and 
          datasetid ne "&EXECUTIONSETID."  and 
          datasetid ne "&PARAMETERSETID."  and 
          datasetid ne "&RULEFIREDSETID.";
       quit;   

       %let ExeNo=&ExeNo.;
       %let ParNo=&ParNo.;
       %let FireNo=&FireNo.;
       %let OutsNo=&OutsNo.; 
       %let byTermCount = 0;

       proc sql noprint;
          select column_properties into :ExeProperty1  - : ExeProperty&ExeNo.
          from &mapping.
          where datasetid="&EXECUTIONSETID.";

          select column_properties into :ParProperty1  - : ParProperty&ParNo.
          from &mapping.
          where datasetid="&PARAMETERSETID.";
           
          select column_properties into :FireProperty1  - : FireProperty&FireNo.
          from &mapping.
          where datasetid="&RULEFIREDSETID.";

          select column_properties into :OutProperty1  - : OutProperty&OutsNo.
          from &mapping.
          where upcase(type)="OUTPUT"           and 
                datasetid ne "&EXECUTIONSETID." and 
                datasetid ne "&PARAMETERSETID." and 
                datasetid ne "&RULEFIREDSETID.";


          * Get list of variables to retain and set to missing.  There is a separate ;
          * set for _N_ = 1 vs first..   Be sure to defer to the _N_=1  section in   ;
          * case a column is in both _N_ = 1 and first.                              ;
          create table work.initRetains as 
             select unique a.termid 
             from work.rulesetoutput as a
             where rulesetid in (
                  select rulesetid from work.rulesetnode
                  where nodeid in (
                           select nodeid 
                           from node 
                           where upcase(section) = 'INIT')) ;

          select count(*) into :byTermCount from ByTerms;

          %if (&byTermCount gt 0) %then %do;    
             select byTerm into :byCols separated by " "
             from ByTerms
             order by bySeqPos;
            
             select byTerm into :finalByTerm
             from ByTerms
             having bySeqPos= max(bySeqPos);


             select b.column  into :rawByCols separated by " "
             from work.ByTerms as a,
                  &mapping  as b
             where a.byTerm = b.termid and upcase(b.type) = 'INPUT'
             order by a.bySeqPos;

             create table work.groupStartRetains as 
                select unique a.termid 
                from work.rulesetoutput as a
                where a.termid not in (select termid from work.initretains) and
                      rulesetid in  (
                          select rulesetid 
                          from work.rulesetnode
                          where nodeid in (
                               select nodeid 
                               from node 
                               where upcase(section) = 'GROUPSTART')) ;
          %end;
       quit;
     
       * Construct a list of terms to issue call missing() on.  These ;
       * are terms that are not in the input data set ;     
       proc contents data=&inputTable
         out=source_vars(keep=name  rename=(name=termid)) noprint;
       run;  

       data work.source_vars ;
         set work.source_vars;
         termid = upcase(termid);
       run;
       proc sort data=work.source_vars; by termid; run;

       data work.allvars;
          set work.rulesetinput(keep=termid)
              work.rulesetoutput(keep=termid);
              termid = upcase(termid);
       run;
       proc sort data=work.allvars nodupkey; by termid; run;

       data work.setmissings;
          set work.allvars(in=_all)
             work.source_vars (in=_source);
          by termid;
          if (_all and first.termid and last.termid) then do;
             output work.setmissings;
          end;
       run;

       %if (&byTermCount gt 0) %then %do;
          %let finalByTerm = %trim(&finalByTerm);
          %let realInputData = work._BRMSorted;
          data _null_;
             file genfile mod LRECL=32767;
             * eventually we want to be smart about sorting. ;
             put "proc sort data= &InputTable out=WORK._BRMSORTED ;";
             put "  by &rawByCols;";
             put " run;";
             put;
          run;
       %end;
       %else %do;
          %let realInputData = &inputTable;
       %end;

       * create the execution and parameter data sets. ;
       data _null_;
          file genfile mod LRECL=32767;
          set  &mapping(keep=termid datasetid 
                       where=(datasetid eq "&EXECUTIONSETID"))
               end=_last;

          if (_N_ eq 1) then do;
              put "data  &&OUTSTATE&EXECUTIONSETID. (keep=";
          end;
          put termid;
          if (_last) then do;
             put ");";
             put;

             %do i=1 %to &ExeNo.;
                put "  &&Exeproperty&i.";
             %end;   
             put "  DEPLMT_EXECUTION_ID=uuidgen();";
             put "  call symput('DEPLMT_EXECUTION_ID',DEPLMT_EXECUTION_ID);";
             put "  DEPLMT_SK=&DeploySk.;";
             put "  DEPLMT_NM='&DeployName.';";
             put "  TRANSACTION_MODE_CD='DIS';";   
             put "  call missing(RECORDS_PROCESSED_NO);";  
             put "  TEST_FLG='N';";    
             put "  START_DTTM=datetime();";
             put "  call missing(END_DTTM);";
             put "  output  &&OUTSTATE&EXECUTIONSETID. ;";
             put "run;";
             put ;
             put "data ";
          end;
       run;

       * Add the keep option. ;
       %do i=1 %to &OutNo.;
          * Ignore the execution and parameter set ids because we only insert into those once. ;
          %if &i. ne &EXECUTIONSETID. and
              &i. ne &PARAMETERSETID. and
              &i. ne &RULEFIREDSETID. %then %do;

             data _null_;
                file genfile mod LRECL=32767;
                set  &mapping(keep=termid datasetid
                      where=(datasetid eq "&&outmap&i"))
                      end=_last;
                if (_n_ eq 1) then do;
                   put "&&OUTSTATE&i. (label=&RuleFlowName keep=";
                end;
                put termid;
                if (_last) then do;
                   put ")";
                   put;
                end;
             run;
          %end;
       %end;
       
        data _null_;
         file genfile mod LRECL=32767;
             %if (%upcase("&RuleFire.") eq "Y") %then %do;
               set  &mapping(keep=termid datasetid
                            where=(datasetid eq "&RULEFIREDSETID"))
                     end=_last;
               if (_n_ eq 1) then do;
                   put "&&OUTSTATE&RULEFIREDSETID (keep=";
               end;
               put termid;
               if (_last) then do;
                  put ")";
                  put;
               end;
             %end;
		run;
		   		 
        
		data _null_;
			file genfile mod LRECL=32767;
			put ';';        
		     
              * Add the output and rule fire column lengths and formats . ;
			  %do i=1 %to &OutsNo.;
              %if "&&OutProperty&i." ne "" %then %do;				
                   put "&&OutProperty&i.";
                %end;
              %end;
              put;

               %if (%upcase("&RuleFire.") eq "Y") %then %do;
                  %do i=1 %to &FireNo.;
                     put "&&FireProperty&i.";
                   %end;
               %end;
               put;
        run;       

       * build call missing() statements ;
       data _null_;
          file genfile mod LRECL=32767;
          %if (%brm_obscount(work.setmissings)) %then %do;
             set work.setmissings end=_last;
             if _n_ eq 1 then do;
                 put 'retain _missingVals 0;';
                 put 'if (not(_missingVals)) then do;';
             end;
             put '    call missing(' termid +(-1) ');';
             if _last then do;
               put '    _missingVals = 1;';
               put 'end;';
               put "set &realInputData (&INPUTTABLEOPTIONS. RENAME=( ";
             end;
          %end;
          %else %do;
             put "set &realInputData (&INPUTTABLEOPTIONS. RENAME=( ";
          %end;
       run;

       * Map the inputs via RENAME= input data set option ;
       data _null_;
          set &Mapping. (where=(upcase(type)="INPUT"));
          file genfile mod LRECL=32767;
          * the pattern is termid = columnname ; 
          if (trim(column) ne trim(termid)) then do;
            assignment=trim(column)||" = "||trim(termid)||" ";
            put assignment; 
          end;
          put;        
       run;

       data _null_;
          file genfile mod LRECL=32767;
          if (_n_ eq 1)  then do;
            put ")) END=_last; ";
            put;
            %if (&byTermCount gt 0) %then %do; 
              put " BY &byCols;";
              put;
            %end;
          end;

          * Map the parameters ;
          set term (where=(upcase(parameter)="TRUE")) end=_eof;
          * the pattern is termid = &termid;
          * for now we will assume DI studio is setting them as macro variables which match the termids. ;
          put termid ' = &' termid ';';           
       
          if (_eof) then do;
            put;
          end;
       run;
    %end; * full code capture;

    * If the ruleset has lookup, then load hash table ;
    %if (&LCount. gt 0) %then %do; 
       proc sql noprint;
          select quote(strip(lookup_id))into :LookupID separated by ','
          from lookups;
       quit;
               
       data _null_;
          file genfile mod LRECL=32767;
          put '/* START OF LOGIC CODE HASH SETUP */';
          put '  retain _hashinit 0;';
          put '  if not(_hashinit) then do;';  

                      
          put '     length LOOKUP_NM $100;';
          put '     length LOOKUP_KEY_NAME_TXT $256;';
          put '     length LOOKUP_VALUE $512;';
         put "     declare hash lookupvalues(dataset:'brm.DCM_SELECTED_LOOKUP_VIEW(where=(LOOKUP_NM in (";
           length lookupstr $32767;
           lookupstr = symget("lookupid");
           look_no=count(lookupstr,',');
           length look $1000;
           do i=1 to look_no+1;
               look=scan(lookupstr,i,',');
               if (i eq look_no+1) then put look;
			   else put look",";
           end;
          put ")))');";
          put "     lookupvalues.DefineKey('LOOKUP_KEY_NAME_TXT');"; 

          put "     lookupvalues.DefineKey('LOOKUP_NM');"; 
                                            
          put "     lookupvalues.DefineData('LOOKUP_VALUE') ;";
          put '     lookupvalues.DefineDone();';
          put '     call missing(LOOKUP_KEY_NAME_TXT,LOOKUP_VALUE, LOOKUP_NM);';
          put '     _hashinit = 1;';
          put '  end;';
          put '/* END OF LOGIC CODE HASH SETUP */';
       run;     
                  
    %end;

    %* Build _N_ eq 1 block start in case there are not retains ;
    data _null_;
      file genfile mod LRECL=32767;
      put 'if (_n_ eq 1) then do;'; 

    run;
    %let currentSection=%str();
    %brs_node(NodeID=&StartNodeID.); 
      

    * the remaining non-logic code generation ; 
        
    %if ("&logicCodeLoc" eq "") %then %do;
       data _null_;
          file genfile mod LRECL=32767;
          * Close off the data step with a RUN statement ;
          put "run;";
          put;
   
          * Store records process number into macro variable ;
          put 'proc sql noprint;';
          put '    select count(*) into :record_process_no';
          put "    from &InputTable.;";
          put 'quit;'; 
          put; 

          * Update the rule execution for number of records processed. ;
          * Also for the completion time. ;
          %if (%upcase("&RuleFire.") eq "Y") %then %do;
             put "data &&OUTSTATE&EXECUTIONSETID.; ";
                         
             put "   set &&OUTSTATE&EXECUTIONSETID. ;";
             put '   RECORDS_PROCESSED_NO=&record_process_no.;';
                 
             put "   END_DTTM=datetime();";
             put "   output &&OUTSTATE&EXECUTIONSETID. ;";
             put "run;";
             put;
          %end;
          * Set QUOTELENMAX back, this is the default setting ;
          put "options QUOTELENMAX;";
          * Release database librefs;
          %if ((&LCount. gt 0) and ("&logicCodeLoc" eq "")) %then %do;
                       
             put '%brm_release;';
          %end;
       run;

 %fastPath:

       * only execute this code if we are not in a requested codegen mode ;
       %include genfile;
          
    %end;

%exit:  
  * Free up the codegen file ;
  %if (%sysfunc(fileref(genfile)) eq 0) %then %do;
     filename genfile clear;
  %end;

  * reset options ;
  options lrecl=&_opt_LRECL  s=&_opt_S s2=&_opt_S2 s2v=&_opt_S2V;

%mend brs_decision_flow;
